# /usr/bin/env python
# pylint: disable=E1101,C0103
# E1101: ignore no member complaints here (created dynamically)
# C0103: also ignore invalid variable name
"""
WMAgent Configuration

Sample WMAgent configuration.
"""
import os
from Utils.PythonVersion import PY3
from WMCore.Configuration import Configuration

# job wrapper script
if PY3:
    submitScript = "etc/submit_py3.sh"
else:
    submitScript = "etc/submit.sh"

# The following parameters may need to be changed.
serverHostName = "HOSTNAME OF WMAGENT MACHINE"
wmbsServicePort = 9997

# The work directory and database need to be separate from the ReqMgr
# installation.
workDirectory = "WMAGENT WORK DIR"
databaseUrl = "oracle://DBUSER:DBPASSWORD@TNSNAME"
databaseUrl = "mysql://DBUSER:DBSPASSWORD@localhost/WMAgentDB"
databaseSocket = "/opt/MySQL-5.1/var/lib/mysql/mysql.sock"

# The couch username and password needs to be added.  The GroupUser and
# ConfigCache couch apps need to be installed into the configcache couch
# database.  The JobDump couchapp needs to be installed into the jobdump
# database.  The GroupUser and ACDC couchapps needs to be install into the
# acdc database.
couchURL = "http://USERNAME:PASSWORD@COUCHSERVER:5984"
logDBName = "wmagent_logdb"
jobDumpDBName = "wmagent_jobdump"
jobSummaryDBName = "wmagent_summary"
summaryStatsDBName = "stat_summary"
acdcDBName = "acdcserver"
workqueueDBName = 'workqueue'
workqueueInboxDbName = 'workqueue_inbox'
# example of workloadSummary url
workloadSummaryDB = "workloadsummary"
workloadSummaryURL = couchURL

# List of BossAir plugins that this agent will use.
bossAirPlugins = ["SimpleCondorPlugin"]

# Required for global pool accounting
glideInAcctGroup = "production"
glideInAcctGroupUser = "cmsdataops"

# DBS Information.
localDBSVersion = "DBS_2_0_8"
globalDBSUrl = "https://cmsweb-prod.cern.ch/dbs/prod/global/DBSReader"
globalDBSVersion = "DBS_2_0_8"

# Job retry information.  How long it will sit in cool off.
retryAlgoParams = {"create": 5000, "submit": 5000, "job": 5000}

# The amount of time to wait after a workflow has completed before archiving it.
workflowArchiveTimeout = 3600

# Global LogLevel
# For setting the general log level of the components
globalLogLevel = 'INFO'

# Contact information
# Used for email alerting
contactName = "cms-service-production-admins@cern.ch"

# Nothing beyond this point should need to be changed.
config = Configuration()

config.section_("Agent")
config.Agent.hostName = serverHostName
config.Agent.contact = contactName
config.Agent.teamName = "REPLACE_TEAM_NAME"
config.Agent.agentName = "WMAgent"
config.Agent.agentNumber = 0
config.Agent.useMsgService = False
config.Agent.useTrigger = False
config.Agent.useHeartbeat = True
config.Agent.isDocker = False

config.section_("General")
config.General.workDir = workDirectory
config.General.logdb_name = logDBName
config.General.central_logdb_url = "need to get from secrets file"
config.General.ReqMgr2ServiceURL = "ReqMgr2 rest service"
config.General.centralWMStatsURL = "Central WMStats URL"
# ReqMgrAux disk cache duration (in hours), set to 5 minutes: 5 / 60 = 0.083
config.General.ReqMgrAuxCacheDuration = 0.083

config.section_("JobStateMachine")
config.JobStateMachine.couchurl = couchURL
config.JobStateMachine.couchDBName = jobDumpDBName
config.JobStateMachine.jobSummaryDBName = jobSummaryDBName
config.JobStateMachine.summaryStatsDBName = summaryStatsDBName
# Amount of documents allowed in the ChangeState module for bulk commits
config.JobStateMachine.maxBulkCommitDocs = 250
# total allowed serialized size for the FJR document that is uploaded to wmagent_jobdump/fwjrs
# NOTE: this needs to be in sync with CouchDB couchdb.max_document_size parameter
# see: https://docs.couchdb.org/en/latest/config/couchdb.html#couchdb/max_document_size
config.JobStateMachine.fwjrLimitSize = 8 * 1000**2  # default: 8 million bytes (not 8MB!!!)

config.section_("ACDC")
config.ACDC.couchurl = "https://cmsweb.cern.ch/couchdb"
config.ACDC.database = acdcDBName

config.section_("WorkloadSummary")
config.WorkloadSummary.couchurl = couchURL
config.WorkloadSummary.database = workloadSummaryDB

config.section_("BossAir")
config.BossAir.pluginDir = "WMCore.BossAir.Plugins"
config.BossAir.pluginNames = bossAirPlugins
config.BossAir.nCondorProcesses = 1
config.BossAir.submitWMSMode = True
config.BossAir.acctGroup = glideInAcctGroup
config.BossAir.acctGroupUser = glideInAcctGroupUser

config.section_("CoreDatabase")
config.CoreDatabase.connectUrl = databaseUrl
# config.CoreDatabase.socket = databaseSocket

config.section_("DashboardReporter")
config.DashboardReporter.dashboardHost = "cms-jobmon.cern.ch"
config.DashboardReporter.dashboardPort = 8884

config.component_('WorkQueueManager')
config.WorkQueueManager.namespace = "WMComponent.WorkQueueManager.WorkQueueManager"
config.WorkQueueManager.componentDir = config.General.workDir + "/WorkQueueManager"
config.WorkQueueManager.level = 'LocalQueue'
config.WorkQueueManager.logLevel = globalLogLevel
config.WorkQueueManager.pollInterval = 180  # 3 min
config.WorkQueueManager.couchurl = couchURL
config.WorkQueueManager.dbname = workqueueDBName
config.WorkQueueManager.inboxDatabase = workqueueInboxDbName
config.WorkQueueManager.rucioUrl = "OVER_WRITE_BY_SECRETS"
config.WorkQueueManager.rucioAuthUrl = "OVER_WRITE_BY_SECRETS"
config.WorkQueueManager.queueParams = {}
config.WorkQueueManager.queueParams["ParentQueueCouchUrl"] = "https://cmsweb.cern.ch/couchdb/workqueue"
# this has to be unique for different work queue. This is just place holder
config.WorkQueueManager.queueParams["QueueURL"] = "http://%s:5984" % (config.Agent.hostName)
config.WorkQueueManager.queueParams["WorkPerCycle"] = 200  # don't pull more than this number of elements per cycle
config.WorkQueueManager.queueParams["QueueDepth"] = 0.5  # pull work from GQ for only half of the resources
# number of available elements to be retrieved within a single CouchDB http request
config.WorkQueueManager.queueParams["RowsPerSlice"] = 2500
# maximum number of available elements rows to be evaluated when acquiring GQ to LQ work
config.WorkQueueManager.queueParams["MaxRowsPerCycle"] = 50000
# Rucio accounts for input data locks and secondary data locks
config.WorkQueueManager.queueParams["rucioAccount"] = "wmcore_transferor"
config.WorkQueueManager.queueParams["rucioAccountPU"] = "wmcore_pileup"


config.component_("DBS3Upload")
config.DBS3Upload.namespace = "WMComponent.DBS3Buffer.DBS3Upload"
config.DBS3Upload.componentDir = config.General.workDir + "/DBS3Upload"
config.DBS3Upload.logLevel = globalLogLevel
config.DBS3Upload.workerThreads = 1
config.DBS3Upload.pollInterval = 100
# "https://cmsweb-prod.cern.ch/dbs/prod/global/DBSWriter" - production one
config.DBS3Upload.dbsUrl = "OVERWRITE_BY_SECRETS"
config.DBS3Upload.primaryDatasetType = "mc"
# provided a block name, this will dump all the block info in a json file
config.DBS3Upload.dumpBlockJsonFor = ""
# set DbsApi requests to use gzip enconding, thus sending compressed data
config.DBS3Upload.gzipEncoding = True
config.DBS3Upload.uploaderName = "WMAgent"

config.section_("DBSInterface")
config.DBSInterface.DBSUrl = globalDBSUrl
config.DBSInterface.DBSVersion = localDBSVersion
config.DBSInterface.globalDBSUrl = globalDBSUrl
config.DBSInterface.globalDBSVersion = globalDBSVersion
config.DBSInterface.MaxFilesToCommit = 200
config.DBSInterface.doGlobalMigration = False
config.DBSInterface.primaryDatasetType = "mc"

config.component_("JobAccountant")
config.JobAccountant.namespace = "WMComponent.JobAccountant.JobAccountant"
config.JobAccountant.componentDir = config.General.workDir + "/JobAccountant"
config.JobAccountant.logLevel = globalLogLevel
config.JobAccountant.workerThreads = 1
config.JobAccountant.pollInterval = 300
config.JobAccountant.specDir = config.General.workDir + "/JobAccountant/SpecCache"

config.component_("JobCreator")
config.JobCreator.namespace = "WMComponent.JobCreator.JobCreator"
config.JobCreator.componentDir = config.General.workDir + "/JobCreator"
config.JobCreator.logLevel = globalLogLevel
config.JobCreator.maxThreads = 1
config.JobCreator.UpdateFromResourceControl = True
config.JobCreator.pollInterval = 120
# This is now OPTIONAL: It defaults to the componentDir
# However: In a production instance, this should be run on a high performance
# disk, and should probably NOT be run on the same disk as the JobArchiver
config.JobCreator.jobCacheDir = config.General.workDir + "/JobCache"
config.JobCreator.defaultJobType = "Processing"
config.JobCreator.workerThreads = 1
# glidein restrictions used for resource estimation (per core)
config.JobCreator.GlideInRestriction = {"MinWallTimeSecs": 1 * 3600,  # 1h
                                        "MaxWallTimeSecs": 45 * 3600,  # pilot lifetime is usually 48h
                                        "MinRequestDiskKB": 1 * 1000 * 1000,  # 1GB
                                        "MaxRequestDiskKB": 20 * 1000 * 1000}  # site limit is ~27GB
config.component_("JobSubmitter")
config.JobSubmitter.namespace = "WMComponent.JobSubmitter.JobSubmitter"
config.JobSubmitter.componentDir = config.General.workDir + "/JobSubmitter"
config.JobSubmitter.logLevel = globalLogLevel
config.JobSubmitter.maxThreads = 1
config.JobSubmitter.pollInterval = 120
config.JobSubmitter.workerThreads = 1
config.JobSubmitter.jobsPerWorker = 100
config.JobSubmitter.maxJobsPerPoll = 1000
config.JobSubmitter.maxJobsToCache = 100000  # used to be 50k
config.JobSubmitter.cacheRefreshSize = 30000  # set -1 if cache need to refresh all the time.
config.JobSubmitter.skipRefreshCount = 20  # (If above the threshold meet, cache will updates every 20 polling cycle) 120 * 20 = 40 minutes
config.JobSubmitter.submitScript = os.path.join(os.environ["WMCORE_ROOT"], submitScript)
config.JobSubmitter.extraMemoryPerCore = 500  # in MB
config.JobSubmitter.drainGraceTime = 2 * 24 * 60 * 60  # in seconds
config.JobSubmitter.useOauthToken = False

config.component_("JobTracker")
config.JobTracker.namespace = "WMComponent.JobTracker.JobTracker"
config.JobTracker.componentDir = config.General.workDir + "/JobTracker"
config.JobTracker.logLevel = globalLogLevel
config.JobTracker.pollInterval = 60

config.component_("JobStatusLite")
config.JobStatusLite.namespace = "WMComponent.JobStatusLite.JobStatusLite"
config.JobStatusLite.componentDir = config.General.workDir + "/JobStatusLite"
config.JobStatusLite.logLevel = globalLogLevel
config.JobStatusLite.pollInterval = 60
config.JobStatusLite.stateTimeouts = {"Error": 300, "Running": 169200, "Pending": 432000}

config.component_("JobUpdater")
config.JobUpdater.namespace = "WMComponent.JobUpdater.JobUpdater"
config.JobUpdater.componentDir = config.General.workDir + "/JobUpdater"
config.JobUpdater.logLevel = globalLogLevel
config.JobUpdater.pollInterval = 120

config.component_("ErrorHandler")
config.ErrorHandler.namespace = "WMComponent.ErrorHandler.ErrorHandler"
config.ErrorHandler.componentDir = config.General.workDir + "/ErrorHandler"
config.ErrorHandler.logLevel = globalLogLevel
config.ErrorHandler.pollInterval = 240
config.ErrorHandler.readFWJR = True
config.ErrorHandler.maxFailTime = 120000
config.ErrorHandler.maxProcessSize = 500

config.component_("RetryManager")
config.RetryManager.namespace = "WMComponent.RetryManager.RetryManager"
config.RetryManager.componentDir = config.General.workDir + "/RetryManager"
config.RetryManager.logLevel = globalLogLevel
config.RetryManager.pollInterval = 240
config.RetryManager.plugins = {"default": "SquaredAlgo"}
config.RetryManager.section_("SquaredAlgo")
config.RetryManager.SquaredAlgo.section_("default")
config.RetryManager.SquaredAlgo.default.coolOffTime = retryAlgoParams

config.component_("JobArchiver")
config.JobArchiver.namespace = "WMComponent.JobArchiver.JobArchiver"
config.JobArchiver.componentDir = config.General.workDir + "/JobArchiver"
config.JobArchiver.pollInterval = 120
config.JobArchiver.logLevel = globalLogLevel
config.JobArchiver.numberOfJobsToCluster = 1000
config.JobArchiver.numberOfJobsToArchive = 10000
# This is now OPTIONAL, it defaults to the componentDir
# HOWEVER: Is is HIGHLY recommended that you do NOT run this on the same
# disk as the JobCreator
# config.JobArchiver.logDir = config.General.workDir + "/JobArchives"

config.component_("TaskArchiver")
config.TaskArchiver.namespace = "WMComponent.TaskArchiver.TaskArchiver"
config.TaskArchiver.componentDir = config.General.workDir + "/TaskArchiver"
config.TaskArchiver.logLevel = globalLogLevel
config.TaskArchiver.pollInterval = 240
config.TaskArchiver.timeOut = workflowArchiveTimeout
config.TaskArchiver.useWorkQueue = True
config.TaskArchiver.workloadSummaryCouchURL = workloadSummaryURL
config.TaskArchiver.workloadSummaryCouchDBName = workloadSummaryDB
config.TaskArchiver.histogramKeys = ["PeakValueRss", "PeakValueVsize", "TotalJobTime", "AvgEventTime"]
config.TaskArchiver.perfPrimaryDatasets = ['SingleMu', 'MuHad', 'MinimumBias']
config.TaskArchiver.perfDashBoardMinLumi = 50
config.TaskArchiver.perfDashBoardMaxLumi = 9000
# dqm address -'https://cmsweb.cern.ch/dqm/dev/'
config.TaskArchiver.dqmUrl = "OVER_WRITE_BY_SECETES"
config.TaskArchiver.requireCouch = True
# set to False couch data if request mgr is not used (Tier0, PromptSkiming)
config.TaskArchiver.useReqMgrForCompletionCheck = True
config.TaskArchiver.localCouchURL = "%s/%s" % (config.JobStateMachine.couchurl,
                                               config.JobStateMachine.couchDBName)
config.TaskArchiver.localQueueURL = "%s/%s" % (config.WorkQueueManager.couchurl,
                                               config.WorkQueueManager.dbname)
config.TaskArchiver.localWMStatsURL = "%s/%s" % (config.JobStateMachine.couchurl,
                                                 config.JobStateMachine.jobSummaryDBName)
config.TaskArchiver.DataKeepDays = 0.125  # couhch history keeping days.
config.TaskArchiver.cleanCouchInterval = 60 * 20  # 20 min
config.TaskArchiver.archiveDelayHours = 24  # delay the archiving so monitor can still show. default 24 hours

# Alert framework configuration

# common 'Alert' section (Alert "senders" use these values to determine destination)
config.section_("Alert")
# destination for the alert messages
config.Alert.address = "tcp://127.0.0.1:6557"
# control channel (internal alert system commands)
config.Alert.controlAddr = "tcp://127.0.0.1:6559"


# mysql*Poller sections were made optional and are defined in the
# wmagent-mod-config file

# Email alert configuration
config.section_("EmailAlert")
config.EmailAlert.toAddr = [contactName] # additional emails can be appended to the list
config.EmailAlert.fromAddr = "noreply@cern.ch"
config.EmailAlert.smtpServer = "localhost"

config.component_("AnalyticsDataCollector")
config.AnalyticsDataCollector.namespace = "WMComponent.AnalyticsDataCollector.AnalyticsDataCollector"
config.AnalyticsDataCollector.componentDir = config.General.workDir + "/AnalyticsDataCollector"
config.AnalyticsDataCollector.logLevel = globalLogLevel
config.AnalyticsDataCollector.pollInterval = 600
config.AnalyticsDataCollector.localCouchURL = "%s/%s" % (config.JobStateMachine.couchurl,
                                                         config.JobStateMachine.couchDBName)
config.AnalyticsDataCollector.localQueueURL = "%s/%s" % (config.WorkQueueManager.couchurl,
                                                         config.WorkQueueManager.dbname)
config.AnalyticsDataCollector.localWMStatsURL = "%s/%s" % (config.JobStateMachine.couchurl,
                                                           config.JobStateMachine.jobSummaryDBName)
config.AnalyticsDataCollector.centralRequestDBURL = "Cental Request DB URL"
config.AnalyticsDataCollector.summaryLevel = "task"
config.AnalyticsDataCollector.couchProcessThreshold = 50
config.AnalyticsDataCollector.pluginName = None

config.component_("ArchiveDataReporter")
config.ArchiveDataReporter.namespace = "WMComponent.ArchiveDataReporter.ArchiveDataReporter"
config.ArchiveDataReporter.componentDir = config.General.workDir + "/ArchiveDataReporter"
config.ArchiveDataReporter.pollInterval = 300
config.ArchiveDataReporter.WMArchiveURL = None
config.ArchiveDataReporter.numDocsRetrievePerPolling = 1000  # number of documents needed to be polled each time
config.ArchiveDataReporter.numDocsUploadPerCall = 200  # number of documents upload each time in bulk to WMArchive

# AgentStatusWatcher has to be the last one in the config to avoid false alarms during startup
config.component_("AgentStatusWatcher")
config.AgentStatusWatcher.namespace = "WMComponent.AgentStatusWatcher.AgentStatusWatcher"
config.AgentStatusWatcher.componentDir = config.General.workDir + "/AgentStatusWatcher"
config.AgentStatusWatcher.logLevel = globalLogLevel
config.AgentStatusWatcher.resourceUpdaterPollInterval = 900  # [second]
config.AgentStatusWatcher.grafanaURL = "https://monit-grafana.cern.ch"
config.AgentStatusWatcher.grafanaToken = "OVERWRITE_BY_SECRETS"
config.AgentStatusWatcher.grafanaSSB = 9475  # monit-grafana API number for Site Status Board
config.AgentStatusWatcher.pendingSlotsSitePercent = 75  # [percent] Pending slots percent over site max running for a site
config.AgentStatusWatcher.pendingSlotsTaskPercent = 70  # [percent] Pending slots percent over task max running for tasks
config.AgentStatusWatcher.runningExpressPercent = 30  # [percent] Only used for tier0 agent
config.AgentStatusWatcher.runningRepackPercent = 10  # [percent] Only used for tier0 agent
config.AgentStatusWatcher.t1SitesCores = 30  # [percent] Only used for tier0 agent
config.AgentStatusWatcher.forceSiteDown = []  # List of sites to be forced to Down status
config.AgentStatusWatcher.onlySSB = False  # Set thresholds for sites only in SSB (Force all other to zero/down)
config.AgentStatusWatcher.enabled = True  # switch to enable or not this component
config.AgentStatusWatcher.agentPollInterval = 300
config.AgentStatusWatcher.drainStatusPollInterval = 3600
config.AgentStatusWatcher.defaultAgentsNumByTeam = 5
config.AgentStatusWatcher.enableAMQ = False
config.AgentStatusWatcher.userAMQ = "OVERWRITE_BY_SECRETS"
config.AgentStatusWatcher.passAMQ = "OVERWRITE_BY_SECRETS"
config.AgentStatusWatcher.topicAMQ = "OVERWRITE_BY_SECRETS"

config.component_("RucioInjector")
config.RucioInjector.namespace = "WMComponent.RucioInjector.RucioInjector"
config.RucioInjector.componentDir = config.General.workDir + "/RucioInjector"
config.RucioInjector.logLevel = globalLogLevel
config.RucioInjector.pollInterval = 300
config.RucioInjector.pollIntervalRules = 43200
config.RucioInjector.cacheExpiration = 2 * 24 * 60 * 60  # two days
config.RucioInjector.createBlockRules = True
config.RucioInjector.RSEPostfix = False  # enable it to append _Test to the RSE names
config.RucioInjector.metaDIDProject = "Production"
config.RucioInjector.containerDiskRuleParams = {"weight": "dm_weight", "copies": 2, "grouping": "DATASET"}
config.RucioInjector.blockRuleParams = {}
# this RSEExpr below might be updated by wmagent-mod-config script
config.RucioInjector.containerDiskRuleRSEExpr = "(tier=2|tier=1)&cms_type=real&rse_type=DISK"
config.RucioInjector.rucioAccount = "OVER_WRITE_BY_SECRETS"
config.RucioInjector.rucioUrl = "OVER_WRITE_BY_SECRETS"
config.RucioInjector.rucioAuthUrl = "OVER_WRITE_BY_SECRETS"

config.component_("WorkflowUpdater")
config.WorkflowUpdater.namespace = "WMComponent.WorkflowUpdater.WorkflowUpdater"
config.WorkflowUpdater.componentDir = config.General.workDir + "/WorkflowUpdater"
config.WorkflowUpdater.logLevel = globalLogLevel
config.WorkflowUpdater.pollInterval = 8 * 60 * 60  # every 8 hours
config.WorkflowUpdater.dbsUrl = "OVER_WRITE_BY_SECRETS"
config.WorkflowUpdater.wmstatsUrl = "OVER_WRITE_BY_SECRETS"
config.WorkflowUpdater.rucioAccount = "wmcore_pileup"
config.WorkflowUpdater.rucioUrl = "OVER_WRITE_BY_SECRETS"
config.WorkflowUpdater.rucioAuthUrl = "OVER_WRITE_BY_SECRETS"
config.WorkflowUpdater.msPileupUrl = "OVER_WRITE_BY_SECRETS"



#!/usr/bin/env python
"""
_JobCountBySubscription_

Monitoring DAO classes for Jobs in WMBS
"""
__all__ = []



from WMCore.WMBS.MySQL.Monitoring.JobCountBySubscriptionAndRun \
  import JobCountBySubscriptionAndRun as JobCountBySubscriptionAndRunMySQL

class JobCountBySubscriptionAndRun(JobCountBySubscriptionAndRunMySQL):
    """
    _JobCountBySubscription_

    return the number of jobs grouped by their status and run for given a subscription (fileset, workflow pair)
    """



#!/usr/bin/env python
"""
Test case for Rucio WMCore Service class
"""
from __future__ import print_function, division, absolute_import

import os
from builtins import range

from future.utils import viewitems
from nose.plugins.attrib import attr
from rucio.client import Client as testClient

from Utils.PythonVersion import PY3
from WMCore.Services.Rucio import Rucio
from WMQuality.Emulators.EmulatedUnitTestCase import EmulatedUnitTestCase

# production/testbed container with 2 blocks
DSET = "/ZeroBias/Run2016B-UL16_ver2_forHarvestOnly-v1/DQMIO"
BLOCK = "/ZeroBias/Run2016B-UL16_ver2_forHarvestOnly-v1/DQMIO#0d25a8ce-ea25-422c-bf86-a37174e7f6a3"
FILE = "/store/data/Run2016B/ZeroBias/DQMIO/UL16_ver2_forHarvestOnly-v1/230000/0C743F08-12F0-D44D-BC39-A740F1F6D623.root"
# production/testbed pileup container with hundreds of blocks (only a few in testbed though)
PUDSET = "/Neutrino_E-10_gun/RunIISummer20ULPrePremix-UL16_106X_mcRun2_asymptotic_v13-v1/PREMIX"
PUBLOCK = "/Neutrino_E-10_gun/RunIISummer20ULPrePremix-UL16_106X_mcRun2_asymptotic_v13-v1/PREMIX#004d5f64-f280-4bad-b88b-4b6e52ea040d"
# production container with 4 blocks
DSET2 = "/Mustar_MuG_L10000_M-3750_TuneCP2_13TeV-pythia8/RunIIFall17NanoAODv6-PU2017_12Apr2018_Nano25Oct2019_102X_mc2017_realistic_v7-v1/NANOAODSIM"
BLOCK2 = "/Mustar_MuG_L10000_M-3750_TuneCP2_13TeV-pythia8/RunIIFall17NanoAODv6-PU2017_12Apr2018_Nano25Oct2019_102X_mc2017_realistic_v7-v1/NANOAODSIM#50aecec1-d5a8-4756-9ee4-f93995c5b524"


# integration containers with with 2 blocks
# DSET3 = "/NoBPTX/Integ_Test-ReReco_LumiMask_HG1812_Validation_TEST_Alan_v13-v11/AOD"
# DSET4 = "/NoBPTX/Integ_Test-ReReco_LumiMask_HG1812_Validation_TEST_Alan_v13-v11/DQMIO"


class RucioTest(EmulatedUnitTestCase):
    """
    Unit tests for Rucio Service module
    """

    def __init__(self, methodName='runTest'):
        # TODO figure out what's going on with CRIC mock
        super(RucioTest, self).__init__(methodName=methodName, mockCRIC=False)

        self.acct = "wma_test"
        self.acct_transf = "wmcore_transferor"

        # HACK: do not verify the SSL certificate because docker images
        # do not contain the CA certificate bundle
        # Relying on the config file in the jenkins infrastructure is a PITA
        # so let's make sure to pass all the necessary arguments
        self.creds = {"client_cert": os.getenv("X509_USER_CERT", "Unknown"),
                      "client_key": os.getenv("X509_USER_KEY", "Unknown")}

        self.defaultArgs = {"host": 'http://cms-rucio-int.cern.ch',
                            "auth_host": 'https://cms-rucio-auth-int.cern.ch',
                            "auth_type": "x509", "account": self.acct,
                            "ca_cert": False, "timeout": 30, "request_retries": 3,
                            "creds": self.creds}

    def setUp(self):
        """
        Setup for unit tests
        """
        super(RucioTest, self).setUp()

        self.myRucio = Rucio.Rucio(self.acct,
                                   hostUrl=self.defaultArgs['host'],
                                   authUrl=self.defaultArgs['auth_host'],
                                   configDict=self.defaultArgs)

        self.client = testClient(rucio_host=self.defaultArgs['host'],
                                 auth_host=self.defaultArgs['auth_host'],
                                 account=self.acct,
                                 ca_cert=self.defaultArgs['ca_cert'],
                                 auth_type=self.defaultArgs['auth_type'],
                                 creds=self.defaultArgs['creds'],
                                 timeout=self.defaultArgs['timeout'])
        if PY3:
            self.assertItemsEqual = self.assertCountEqual

    def tearDown(self):
        """
        Nothing to be done for this case
        """
        pass

    def createProductionRucio(self, account=None):
        """Create an object pointing to the Rucio Production instance"""
        # NOTE: do your best to avoid testing against production!!!
        newParams = {"auth_type": "x509", "ca_cert": False, "timeout": 50}
        prodRucio = Rucio.Rucio(account,
                                hostUrl='http://cms-rucio.cern.ch',
                                authUrl='https://cms-rucio-auth.cern.ch',
                                configDict=newParams)
        return prodRucio

    def testConfig(self):
        """
        Test service attributes and the override mechanism
        """
        for key in self.defaultArgs:
            self.assertEqual(getattr(self.myRucio.cli, key), self.defaultArgs[key])
        self.assertTrue(getattr(self.myRucio.cli, "user_agent").startswith("wmcore-client/"))
        self.assertTrue(getattr(self.client, "user_agent").startswith("rucio-clients/"))

        newParams = {"host": 'http://cms-rucio-int.cern.ch',
                     "auth_host": 'https://cms-rucio-auth-int.cern.ch',
                     "auth_type": "x509", "account": self.acct,
                     "ca_cert": False, "timeout": 5}
        newKeys = list(newParams)

        rucio = Rucio.Rucio(newParams['account'], hostUrl=newParams['host'],
                            authUrl=newParams['auth_host'], configDict=newParams)

        for key in newKeys:
            self.assertEqual(getattr(rucio.cli, key), newParams[key])

    def testGetAccount(self):
        """
        Test whether we can fetch data about a specific rucio account
        """
        res = self.client.get_account(self.acct)
        res2 = self.myRucio.getAccount(self.acct)
        self.assertEqual(res['account'], self.acct)
        self.assertEqual(res['status'], "ACTIVE")
        self.assertEqual(res['account_type'], "USER")
        self.assertTrue({"status", "account", "account_type"}.issubset(set(res2.keys())))
        self.assertTrue({self.acct, "ACTIVE", "USER"}.issubset(set(res2.values())))

    def testGetAccountUsage(self):
        """
        Test whether we can fetch the data usage for a given rucio account
        """
        # test against a specific RSE
        res = list(self.client.get_local_account_usage(self.acct, rse="T1_US_FNAL_Disk"))
        res2 = self.myRucio.getAccountUsage(self.acct, rse="T1_US_FNAL_Disk")
        self.assertEqual(res[0]["rse"], "T1_US_FNAL_Disk")
        self.assertTrue(res[0]["files"] >= 1)
        self.assertTrue(res[0]["bytes"] >= 1000)
        self.assertTrue("bytes_remaining" in res[0])
        self.assertTrue(res[0]["bytes_limit"] >= 0)
        self.assertEqual(res, res2)

        # test against all RSEs
        res = list(self.client.get_local_account_usage(self.acct))
        res2 = self.myRucio.getAccountUsage(self.acct)
        self.assertTrue(len(res) > 1)
        self.assertTrue(len(res2) > 1)
        # I have manually created a rule for this account, so it will be there...
        self.assertEqual(res, res2)

        # now test against an account that either does not exist or that we cannot access
        res = self.myRucio.getAccountUsage("admin")
        self.assertIsNone(res)

    def testGetAccountLimits(self):
        """
        Test whether we can fetch the data quota for a given rucio account
        """
        res = self.client.get_local_account_limits(self.acct_transf)
        res2 = self.myRucio.getAccountLimits(self.acct_transf)
        # {'T2_CH_CERN': 10000000000000, 'T2_US_Purdue': 5000000000000, 'T1_US_FNAL_Disk': 10000000000000}
        self.assertTrue(len(res) >= 3)
        self.assertTrue(len(res2) >= 3)
        for rse in ('T1_US_FNAL_Disk', 'T2_CH_CERN', 'T2_US_Purdue'):
            self.assertEqual(res[rse], res2[rse])

        # test an account that we have no access to
        res = self.myRucio.getAccountLimits("wma_prod")
        self.assertEqual(res, {})

        # finally, test an account that does not exist
        res = self.myRucio.getAccountLimits("any_random_account")
        self.assertEqual(res, {})

    # @attr('integration')
    def testWhoAmI(self):
        """
        Test user mapping information from the request headers
        """
        res = dict(self.client.whoami())
        res2 = dict(self.myRucio.whoAmI())
        self.assertTrue({"status", "account"}.issubset(set(res.keys())))
        self.assertTrue(set(res.keys()) == set(res2.keys()))

    def testPing(self):
        """
        Tests server ping
        """
        res = self.client.ping()
        res2 = self.myRucio.pingServer()
        self.assertTrue("version" in res)
        self.assertItemsEqual(res, res2)

    def testGetBlocksInContainer(self):
        """
        Test `getBlocksInContainer` method, the ability to retrieve blocks
        inside a container.
        """
        # test a CMS dataset that does not exist
        with self.assertRaises(Rucio.WMRucioDIDNotFoundException):
            self.myRucio.getBlocksInContainer("Alan")

        # provide a CMS block instead of a dataset
        res = self.myRucio.getBlocksInContainer(BLOCK)
        self.assertEqual(res, [])

        # finally provide a real CMS dataset
        res = self.myRucio.getBlocksInContainer(DSET)
        self.assertTrue(len(res) >= len([BLOCK]))
        self.assertIn(BLOCK, res)

    def testGetReplicaInfoForBlocks(self):
        """
        Test `getReplicaInfoForBlocks` method.
        """
        res = self.myRucio.getReplicaInfoForBlocks(dataset=DSET)
        self.assertTrue(isinstance(res, list))
        self.assertTrue(len(res) >= 1)  # at this very moment, there are 11 replicas
        blocks = [item['name'] for item in res]
        self.assertTrue(BLOCK in blocks)
        for item in res:
            self.assertTrue(len(item['replica']) > 0)

        # Setting  deep=True should yield the same results
        res = self.myRucio.getReplicaInfoForBlocks(dataset=DSET, deep=True)
        self.assertTrue(isinstance(res, list))
        self.assertTrue(len(res) >= 1)  # Again, there are 11 replicas
        blocks = [item['name'] for item in res]
        self.assertTrue(BLOCK in blocks)
        for item in res:
            self.assertTrue(len(item['replica']) > 0)

    def testGetPFN(self):
        """
        Test `getPFN` method
        """
        ### New endpoint for Rucio Int
        cernTestbedPrefix = 'gsiftp://eoscmsftp.cern.ch:2811/eos/cms/store/test/rucio/int'

        testLfn = "/store"
        resp = self.myRucio.getPFN(site="T2_CH_CERN", lfns=testLfn)
        # self.assertEqual(resp[testLfn], cernTestDefaultPrefix + testLfn)
        self.assertEqual(resp[testLfn], cernTestbedPrefix + testLfn)

        # we do not rely on the following check with lfns=""
        # but since it currently works, it will be nice that it keeps working when
        # site TFCs will be replaced by prefixes and this may the only. Beware that currently
        # it only works for a subset of sites, as it relies on the specifics of their TFC.
        resp = self.myRucio.getPFN(site="T2_CH_CERN", lfns="")
        self.assertEqual(resp, {'': cernTestbedPrefix + "/"})

        # possible additional tests (from Stefano)
        lfn1 = '/store/afile'
        lfn2 = '/store/mc/afile'
        resp = self.myRucio.getPFN(site="T2_CH_CERN", lfns=lfn1)
        # self.assertEqual(resp[lfn1], cernTestDefaultPrefix + lfn1)
        self.assertEqual(resp[lfn1], cernTestbedPrefix + lfn1)

        # test with a list of LFN's
        resp = self.myRucio.getPFN(site="T2_CH_CERN", lfns=[lfn1, lfn2])
        self.assertEqual(len(resp), 2)
        # self.assertEqual(resp[lfn1], cernTestDefaultPrefix + lfn1)
        self.assertEqual(resp[lfn1], cernTestbedPrefix + lfn1)
        # self.assertEqual(resp[lfn2], cernTestDefaultPrefix + lfn2)
        self.assertEqual(resp[lfn2], cernTestbedPrefix + lfn2)

        # test different protocols
        resp = self.myRucio.getPFN(site="T2_US_Nebraska", lfns=lfn1, protocol='gsiftp')
        self.assertEqual(resp[lfn1],
                         "gsiftp://red-gridftp.unl.edu:2811/mnt/hadoop/user/uscms01/pnfs/unl.edu/data4/cms/store/test/rucio/int" + lfn1)
        # resp = self.myRucio.getPFN(site="T2_US_Nebraska", lfns=lfn1, protocol='xrootd')
        # self.assertEqual(resp[lfn1], "davs://xrootd-local.unl.edu:1094/store/test/rucio/int" + lfn1)

    @attr('integration')
    def testProdGetPFN(self):
        """
        Test `getPFN` method using the production server, hence set
        as an integration test not to be executed by jenkins
        """
        prodRucio = self.createProductionRucio("wmcore_transferor")
        # simplest call (from Alan)
        cernTestDefaultPrefix = "gsiftp://eoscmsftp.cern.ch:2811/eos/cms"
        testLfn = "/store"
        resp = prodRucio.getPFN(site="T2_CH_CERN", lfns=testLfn)
        self.assertEqual(resp[testLfn], cernTestDefaultPrefix + testLfn)

        # see comment in testGetPFN() function above about testing with lfns=""
        resp = prodRucio.getPFN(site="T2_CH_CERN", lfns="")
        self.assertItemsEqual(resp, {u'': u'gsiftp://eoscmsftp.cern.ch:2811/'})

        # possible additional tests (from Stefano)
        lfn1 = '/store/afile'
        lfn2 = '/store/mc/afile'
        resp = prodRucio.getPFN(site="T2_CH_CERN", lfns=lfn1)
        self.assertEqual(resp[lfn1], cernTestDefaultPrefix + lfn1)

        # test with a list of LFN's
        resp = prodRucio.getPFN(site="T2_CH_CERN", lfns=[lfn1, lfn2])
        self.assertEqual(len(resp), 2)
        self.assertEqual(resp[lfn1], cernTestDefaultPrefix + lfn1)
        self.assertEqual(resp[lfn2], cernTestDefaultPrefix + lfn2)

        # test different protocols
        resp = prodRucio.getPFN(site="T2_US_Nebraska", lfns=lfn1, protocol='gsiftp')
        self.assertEqual(resp[lfn1],
                         "gsiftp://red-gridftp.unl.edu:2811/mnt/hadoop/user/uscms01/pnfs/unl.edu/data4/cms" + lfn1)
        resp = prodRucio.getPFN(site="T2_US_Nebraska", lfns=lfn1, protocol='davs')
        self.assertEqual(resp[lfn1], "davs://xrootd-local.unl.edu:1094" + lfn1)

    def testListContent(self):
        """
        Test `listContent` method, to list content of a given DID
        """
        # listing blocks for a dataset
        res = self.myRucio.listContent(DSET)
        self.assertTrue(len(res) >= 2)
        self.assertEqual(res[0]["type"], "DATASET")

        # listing files for a block
        res = self.myRucio.listContent(BLOCK)
        self.assertTrue(len(res) > 5)
        self.assertEqual(res[0]["type"], "FILE")

        res = self.myRucio.listContent("/Primary/ProcStr-v1/tier")
        self.assertItemsEqual(res, [])

    def testListDataRules(self):
        """
        Test `listDataRules` method
        """
        res = self.myRucio.listDataRules(DSET)
        self.assertItemsEqual(res, [])

    @attr('integration')
    def testListDataRules2(self):
        """
        Test `listDataRules` method with data from production
        """
        prodRucio = self.createProductionRucio("wmcore_transferor")

        resp = prodRucio.listDataRules(name=DSET2, account="transfer_ops")
        self.assertEqual(len(resp), 5)
        accts = set([rule['account'] for rule in resp])
        self.assertItemsEqual(accts, ['transfer_ops'])

        resp = prodRucio.listDataRules(name=DSET2, account="wmcore_transferor")
        self.assertItemsEqual(resp, [])

    def testGetRule(self):
        """
        Test `getRule` method
        """
        # Badly formatted rule id, raises/catches a general exception
        res = self.myRucio.getRule("blah")
        self.assertItemsEqual(res, {})

        # Properly formatted rule, but inexistent id
        res = self.myRucio.getRule("1d6ea1d916d5492e81b1bb30ed4aebc0")
        self.assertItemsEqual(res, {})

        # Properly formatted rule, rule manually created
        res = self.myRucio.getRule("eb4448a33f8f4c3e8d162be42e7a2eb1")
        self.assertTrue(res)

    def testListParentDIDs(self):
        """
        Test `listParentDIDs` method, which lists the parent DIDs
        """
        # given a file, list its blocks (for now, single block)
        res = self.myRucio.listParentDIDs(FILE)
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0]['scope'], "cms")
        self.assertEqual(res[0]['type'], "DATASET")
        self.assertEqual(res[0]['name'], BLOCK)

        # given a block, list its containers (for now, single container)
        res = self.myRucio.listParentDIDs(BLOCK)
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0]['scope'], "cms")
        self.assertEqual(res[0]['type'], "CONTAINER")
        self.assertEqual(res[0]['name'], DSET)

        # given a container, returns nothing because containers have no parents
        res = self.myRucio.listParentDIDs(DSET)
        self.assertEqual(res, [])

    def testEvaluateRSEExpression(self):
        """
        Test the `evaluateRSEExpression` method
        """
        for i in range(2):
            res = self.myRucio.evaluateRSEExpression("tier=1", useCache=False)
            self.assertTrue(len(res) > 1)
            self.assertTrue("T1_US_FNAL_Disk" in res)
        self.myRucio.evaluateRSEExpression("tier=1", useCache=True)

    def testPickRSE(self):
        """
        Test the `pickRSE` method
        """
        print(self.myRucio.cli.list_rse_attributes("T1_US_FNAL_Disk"))
        resp = self.myRucio.pickRSE(rseExpression="dm_weight>0", rseAttribute="dm_weight")
        self.assertTrue(len(resp) == 2)
        self.assertTrue(resp[1] is True or resp[1] is False)

    def testRequiresApproval(self):
        """
        Test the `pickRSE` method
        """
        # NOTE: these are sort of test RSEs in Rucio Int
        resp = self.myRucio.requiresApproval("T1_US_FNAL_Tape_Input")
        self.assertTrue(resp)
        resp = self.myRucio.requiresApproval("T1_IT_CNAF_Tape_Input")
        self.assertFalse(resp)

    @attr('integration')  # jenkins cannot access this rucio account
    def testGetPileupLockedAndAvailable(self):
        """
        Test `getPileupLockedAndAvailable` method
        """
        # as much as I dislike it, we need to use the production instance...
        prodRucio = self.createProductionRucio("wmcore_transferor")
        resp = prodRucio.getPileupLockedAndAvailable(PUDSET, "transfer_ops")
        # this dataset contains 10 blocks
        self.assertTrue(len(resp) > 100)
        self.assertTrue(PUBLOCK in resp)
        # with more than 10 block replicas in the grid
        for block, rses in viewitems(resp):
            self.assertTrue(len(rses) >= 2)

    @attr('integration')  # jenkins cannot access this rucio account
    def testGetDataLockedAndAvailable(self):
        """
        Test `getDataLockedAndAvailable` method
        """
        # as much as I dislike it, we need to use the production instance...
        prodRucio = self.createProductionRucio("wmcore_transferor")

        # This is a very heavy check, with ~25k blocks
        # PUDSET = "/Neutrino_E-10_gun/RunIISpring15PrePremix-PUMoriond17_80X_mcRun2_asymptotic_2016_TrancheIV_v2-v2/GEN-SIM-DIGI-RAW"
        expectedRses = ['T2_CH_CERN', 'T1_US_FNAL_Disk', 'T1_UK_RAL_Disk', 'T1_UK_RAL_Tape']
        resp = prodRucio.getDataLockedAndAvailable(name=PUDSET)
        self.assertItemsEqual(resp, expectedRses)

        # matches rules for this container and transfer_ops account, returns an union of the blocks RSEs
        resp = prodRucio.getDataLockedAndAvailable(name=DSET2, account="transfer_ops")
        self.assertItemsEqual(resp, ['T2_CH_CSCS', 'T1_FR_CCIN2P3_Disk', 'T1_IT_CNAF_Disk', 'T2_US_Purdue',
                                     'T1_RU_JINR_Disk', 'T2_FI_HIP', 'T2_UK_London_IC', 'T2_FR_GRIF_LLR',
                                     'T2_US_Nebraska', 'T2_IT_Bari', 'T2_FR_IPHC', 'T2_DE_DESY', 'T2_BE_IIHE',
                                     'T1_DE_KIT_Disk', 'T2_BR_SPRACE', 'T1_US_FNAL_Disk', 'T1_UK_RAL_Disk',
                                     'T2_US_Vanderbilt'])

        # matches the same rules as above (all multi RSEs), but returns an intersection of the RSEs
        resp = prodRucio.getDataLockedAndAvailable(name=DSET2, account="transfer_ops", grouping="ALL")
        # FIXME: there is a serious block distribution problem with Rucio, there should be 6 disk RSE locations..
        self.assertItemsEqual(resp, ['T1_US_FNAL_Disk'])

        # there are no rules for the blocks, but 6 copies for the container level
        resp = prodRucio.getDataLockedAndAvailable(name=BLOCK2, account="transfer_ops", grouping="ALL")
        self.assertItemsEqual(resp, ['T1_IT_CNAF_Disk', 'T1_RU_JINR_Disk', 'T1_UK_RAL_Disk', 'T1_US_FNAL_Disk',
                                     'T2_IT_Bari', 'T2_US_Vanderbilt'])

        # return the  6 copies for the container level, plus the tape one(s)
        resp = prodRucio.getDataLockedAndAvailable(name=BLOCK2, account="transfer_ops",
                                                   grouping="ALL", returnTape=True)
        self.assertItemsEqual(resp, ['T1_IT_CNAF_Disk', 'T1_RU_JINR_Disk', 'T1_UK_RAL_Disk', 'T1_US_FNAL_Disk',
                                     'T2_IT_Bari', 'T2_US_Vanderbilt', 'T1_IT_CNAF_Tape'])

        # there are no rules for the blocks, but 6 copies for the container level
        resp = prodRucio.getDataLockedAndAvailable(name=BLOCK2, account="transfer_ops", grouping="DATASET")
        self.assertItemsEqual(resp, [])



#!/usr/bin/env python
"""
_InsertWorkflow_

Oracle implementation of DBSBuffer.InsertWorkflow
"""

from WMComponent.DBS3Buffer.MySQL.InsertWorkflow import InsertWorkflow as MySQLInsertWorkflow

class InsertWorkflow(MySQLInsertWorkflow):
    """
    _InsertWorkflow_

    Insert a workflow using the name and task
    """

    sql = """INSERT INTO dbsbuffer_workflow
                (id, name, task,
                block_close_max_wait_time,
                block_close_max_files,
                block_close_max_events,
                block_close_max_size)
             SELECT dbsbuffer_workflow_seq.nextval,
                    :name,
                    :task,
                    :blockMaxCloseTime,
                    :blockMaxFiles,
                    :blockMaxEvents,
                    :blockMaxSize
             FROM DUAL
             WHERE NOT EXISTS (
               SELECT id
               FROM dbsbuffer_workflow
               WHERE name = :name
               AND task = :task
             )"""



#! /usr/bin/env python

from builtins import str, bytes

import subprocess
import os
import re
import zlib
import base64
import sys
from types import ModuleType, FunctionType
from gc import get_referents

def lowerCmsHeaders(headers):
    """
    Lower CMS headers in provided header's dict. The WMCore Authentication
    code check only cms headers in lower case, e.g. cms-xxx-yyy.
    """
    lheaders = {}
    for hkey, hval in list(headers.items()): # perform lower-case
        # lower header keys since we check lower-case in headers
        if hkey.startswith('Cms-') or hkey.startswith('CMS-'):
            lheaders[hkey.lower()] = hval
        else:
            lheaders[hkey] = hval
    return lheaders

def makeList(stringList):
    """
    _makeList_

    Make a python list out of a comma separated list of strings,
    throws a ValueError if the input is not well formed.
    If the stringList is already of type list, then return it untouched.
    """
    if isinstance(stringList, list):
        return stringList
    if isinstance(stringList, str):
        toks = stringList.lstrip(' [').rstrip(' ]').split(',')
        if toks == ['']:
            return []
        return [str(tok.strip(' \'"')) for tok in toks]
    raise ValueError("Can't convert to list %s" % stringList)


def makeNonEmptyList(stringList):
    """
    _makeNonEmptyList_

    Given a string or a list of strings, return a non empty list of strings.
    Throws an exception in case the final list is empty or input data is not
    a string or a python list
    """
    finalList = makeList(stringList)
    if not finalList:
        raise ValueError("Input data cannot be an empty list %s" % stringList)
    return finalList


def strToBool(string):
    """
    Try to convert different variations of True or False (including a string
    type object) to a boolean value.
    In short:
     * True gets mapped from: True, "True", "true", "TRUE".
     * False gets mapped from: False, "False", "false", "FALSE"
     * anything else will fail
    :param string: expects a boolean or a string, but it could be anything else
    :return: a boolean value, or raise an exception if value passed in is not supported
    """
    if string is False or string is True:
        return string
    elif string in ["True", "true", "TRUE"]:
        return True
    elif string in ["False", "false", "FALSE"]:
        return False
    raise ValueError("Can't convert to bool: %s" % string)


def safeStr(string):
    """
    _safeStr_

    Cast simple data (int, float, basestring) to string.
    """
    if not isinstance(string, (tuple, list, set, dict)):
        return str(string)
    raise ValueError("We're not supposed to convert %s to string." % string)


def diskUse():
    """
    This returns the % use of each disk partition
    """
    diskPercent = []
    df = subprocess.Popen(["df", "-klP"], stdout=subprocess.PIPE)
    output = df.communicate()[0]
    output = decodeBytesToUnicode(output).split("\n")
    for x in output:
        split = x.split()
        if split != [] and split[0] != 'Filesystem':
            diskPercent.append({'filesystem': split[0],
                                'mounted':    split[5], 
                                'percent':    split[4]})

    return diskPercent


def numberCouchProcess():
    """
    This returns the number of couch process
    """
    ps = subprocess.Popen(["ps", "-ef"], stdout=subprocess.PIPE)
    process = ps.communicate()[0]
    process = decodeBytesToUnicode(process).count('couchjs')

    return process


def rootUrlJoin(base, extend):
    """
    Adds a path element to the path within a ROOT url
    """
    if base:
        match = re.match("^root://([^/]+)/(.+)", base)
        if match:
            host = match.group(1)
            path = match.group(2)
            newpath = os.path.join(path, extend)
            newurl = "root://%s/%s" % (host, newpath)
            return newurl
    return None


def zipEncodeStr(message, maxLen=5120, compressLevel=9, steps=100, truncateIndicator=" (...)"):
    """
    _zipEncodeStr_
    Utility to zip a string and encode it.
    If zipped encoded length is greater than maxLen,
    truncate message until zip/encoded version
    is within the limits allowed.
    """
    message = encodeUnicodeToBytes(message)
    encodedStr = zlib.compress(message, compressLevel)
    encodedStr = base64.b64encode(encodedStr)
    if len(encodedStr) < maxLen or maxLen == -1:
        return encodedStr

    compressRate = 1. * len(encodedStr) / len(base64.b64encode(message))

    # Estimate new length for message zip/encoded version
    # to be less than maxLen.
    # Also, append truncate indicator to message.
    truncateIndicator = encodeUnicodeToBytes(truncateIndicator)
    strLen = int((maxLen - len(truncateIndicator)) / compressRate)
    message = message[:strLen] + truncateIndicator

    encodedStr = zipEncodeStr(message, maxLen=-1)

    # If new length is not short enough, truncate
    # recursively by steps
    while len(encodedStr) > maxLen:
        message = message[:-steps - len(truncateIndicator)] + truncateIndicator
        encodedStr = zipEncodeStr(message, maxLen=-1)

    return encodedStr


def getSize(obj):
    """
    _getSize_

    Function to traverse an object and calculate its total size in bytes
    :param obj: a python object
    :return: an integer representing the total size of the object

    Code extracted from Stack Overflow:
    https://stackoverflow.com/questions/449560/how-do-i-determine-the-size-of-an-object-in-python
    """
    # Custom objects know their class.
    # Function objects seem to know way too much, including modules.
    # Exclude modules as well.
    BLACKLIST = type, ModuleType, FunctionType

    if isinstance(obj, BLACKLIST):
        raise TypeError('getSize() does not take argument of type: '+ str(type(obj)))
    seen_ids = set()
    size = 0
    objects = [obj]
    while objects:
        need_referents = []
        for obj in objects:
            if not isinstance(obj, BLACKLIST) and id(obj) not in seen_ids:
                seen_ids.add(id(obj))
                size += sys.getsizeof(obj)
                need_referents.append(obj)
        objects = get_referents(*need_referents)
    return size


def decodeBytesToUnicode(value, errors="strict"):
    """
    Accepts an input "value" of generic type.

    If "value" is a string of type sequence of bytes (i.e. in py2 `str` or
    `future.types.newbytes.newbytes`, in py3 `bytes`), then it is converted to
    a sequence of unicode codepoints.

    This function is useful for cleaning input data when using the
    "unicode sandwich" approach, which involves converting bytes (i.e. strings
    of type sequence of bytes) to unicode (i.e. strings of type sequence of
    unicode codepoints, in py2 `unicode` or `future.types.newstr.newstr`,
    in py3 `str` ) as soon as possible when recieving input data, and
    converting unicode back to bytes as late as possible.
    achtung!:
    - converting unicode back to bytes is not covered by this function
    - converting unicode back to bytes is not always necessary. when in doubt,
      do not do it.
    Reference: https://nedbatchelder.com/text/unipain.html

    py2:
    - "errors" can be: "strict", "ignore", "replace",
    - ref: https://docs.python.org/2/howto/unicode.html#the-unicode-type
    py3:
    - "errors" can be: "strict", "ignore", "replace", "backslashreplace"
    - ref: https://docs.python.org/3/howto/unicode.html#the-string-type
    """
    if isinstance(value, bytes):
        return value.decode("utf-8", errors)
    return value

def decodeBytesToUnicodeConditional(value, errors="ignore", condition=True):
    """
    if *condition*, then call decodeBytesToUnicode(*value*, *errors*),
    else return *value*
    
    This may be useful when we want to conditionally apply decodeBytesToUnicode,
    maintaining brevity.

    Parameters
    ----------
    value : any
        passed to decodeBytesToUnicode
    errors: str
        passed to decodeBytesToUnicode
    condition: boolean of object with attribute __bool__()
        if True, then we run decodeBytesToUnicode. Usually PY2/PY3
    """
    if condition:
        return decodeBytesToUnicode(value, errors)
    return value

def encodeUnicodeToBytes(value, errors="strict"):
    """
    Accepts an input "value" of generic type.

    If "value" is a string of type sequence of unicode (i.e. in py2 `unicode` or
    `future.types.newstr.newstr`, in py3 `str`), then it is converted to
    a sequence of bytes.

    This function is useful for encoding output data when using the
    "unicode sandwich" approach, which involves converting unicode (i.e. strings
    of type sequence of unicode codepoints) to bytes (i.e. strings of type
    sequence of bytes, in py2 `str` or `future.types.newbytes.newbytes`,
    in py3 `bytes`) as late as possible when passing a string to a third-party
    function that only accepts bytes as input (pycurl's curl.setop is an
    example).
    py2:
    - "errors" can be: "strict", "ignore", "replace", "xmlcharrefreplace"
    - ref: https://docs.python.org/2/howto/unicode.html#the-unicode-type
    py3:
    - "errors" can be: "strict", "ignore", "replace", "backslashreplace", 
      "xmlcharrefreplace", "namereplace"
    - ref: https://docs.python.org/3/howto/unicode.html#the-string-type
    """
    if isinstance(value, str):
        return value.encode("utf-8", errors)
    return value

def encodeUnicodeToBytesConditional(value, errors="ignore", condition=True):
    """
    if *condition*, then call encodeUnicodeToBytes(*value*, *errors*),
    else return *value*
    
    This may be useful when we want to conditionally apply encodeUnicodeToBytes,
    maintaining brevity.

    Parameters
    ----------
    value : any
        passed to encodeUnicodeToBytes
    errors: str
        passed to encodeUnicodeToBytes
    condition: boolean of object with attribute __bool__()
        if True, then we run encodeUnicodeToBytes. Usually PY2/PY3
    """
    if condition:
        return encodeUnicodeToBytes(value, errors)
    return value

def normalize_spaces(text):
    """
    Helper function to remove any number of empty spaces within given text and replace
    then with single space.
    :param text: string
    :return: normalized string
    """
    return re.sub(r'\s+', ' ', text).strip()



#!/usr/bin/env python
"""
_Job_t_

Unit tests for the WMBS job class.
"""
from __future__ import absolute_import

from builtins import str
import threading
import unittest

from WMCore.DataStructs.Run import Run
from WMCore.Services.UUIDLib import makeUUID
from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset as Fileset
from WMCore.WMBS.Job import Job
from WMCore.WMBS.JobGroup import JobGroup
from WMCore.WMBS.Mask import Mask
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.WorkUnit import WorkUnit
from WMCore.WMBS.Workflow import Workflow
from WMCore_t.WMBS_t.JobTestBase import JobTestBase


class JobTest(JobTestBase):
    def testCreateDeleteExists(self):
        """
        _testCreateDeleteExists_

        Create and then delete a job.  Use the job class's exists() method to
        determine if the job has been written to the database before it is
        created, after it has been created and after it has been deleted.
        """
        testWorkflow = Workflow(spec="spec.xml", owner="Simon",
                                name="wf001", task="Test")
        testWorkflow.create()

        testWMBSFileset = Fileset(name="TestFileset")
        testWMBSFileset.create()

        testSubscription = Subscription(fileset=testWMBSFileset,
                                        workflow=testWorkflow)
        testSubscription.create()

        testJobGroup = JobGroup(subscription=testSubscription)
        testJobGroup.create()

        testFileA = File(lfn="/this/is/a/lfnA", size=1024, events=10)
        testFileB = File(lfn="/this/is/a/lfnB", size=1024, events=10)
        testFileA.create()
        testFileB.create()

        testJob = Job(name="TestJob", files=[testFileA, testFileB])
        # testWU = WorkUnit(taskID=testWorkflow.id, fileid=testFileA['id'], runLumi=Run(1, *[44]))

        self.assertFalse(testJob.exists(), "Job exists before it was created")
        # self.assertFalse(testWU.exists(), "WorkUnit exists before it was created")

        testJob.create(group=testJobGroup)

        self.assertTrue(testJob.exists(), "Job does not exist after it was created")
        # self.assertFalse(testWU.exists(), "WorkUnit exists when there is no work")

        # Test the getWorkflow method
        workflow = testJob.getWorkflow()
        self.assertEqual(workflow['task'], 'Test')
        self.assertEqual(workflow['name'], 'wf001')

        testJob.delete()

        self.assertFalse(testJob.exists(), "Job exists after it was deleted")
        # self.assertFalse(testWU.exists(), "WorkUnit exists after job is deleted")

        return

    def testCreateTransaction(self):
        """
        _testCreateTransaction_

        Create a job and save it to the database.  Roll back the database
        transaction and verify that the job is no longer in the database.
        """
        testWorkflow = Workflow(spec="spec.xml", owner="Simon",
                                name="wf001", task="Test")
        testWorkflow.create()

        testWMBSFileset = Fileset(name="TestFileset")
        testWMBSFileset.create()

        testSubscription = Subscription(fileset=testWMBSFileset,
                                        workflow=testWorkflow)
        testSubscription.create()

        testJobGroup = JobGroup(subscription=testSubscription)
        testJobGroup.create()

        testFileA = File(lfn="/this/is/a/lfnA", size=1024, events=10)
        testFileB = File(lfn="/this/is/a/lfnB", size=1024, events=10)
        testFileA.create()
        testFileB.create()

        myThread = threading.currentThread()
        myThread.transaction.begin()

        testJob = Job(name="TestJob", files=[testFileA, testFileB])

        assert testJob.exists() is False, \
            "ERROR: Job exists before it was created"

        testJob.create(group=testJobGroup)

        assert testJob.exists() >= 0, \
            "ERROR: Job does not exist after it was created"

        myThread.transaction.rollback()

        assert testJob.exists() is False, \
            "ERROR: Job exists after transaction was rolled back."

        return

    def testDeleteTransaction(self):
        """
        _testDeleteTransaction_

        Create a new job and commit it to the database.  Start a new transaction
        and delete the file from the database.  Verify that the file has been
        deleted.  After that, roll back the transaction and verify that the
        job is once again in the database.
        """
        testWorkflow = Workflow(spec="spec.xml", owner="Simon",
                                name="wf001", task="Test")
        testWorkflow.create()

        testWMBSFileset = Fileset(name="TestFileset")
        testWMBSFileset.create()

        testSubscription = Subscription(fileset=testWMBSFileset,
                                        workflow=testWorkflow)
        testSubscription.create()

        testJobGroup = JobGroup(subscription=testSubscription)
        testJobGroup.create()

        testFileA = File(lfn="/this/is/a/lfnA", size=1024, events=10)
        testFileB = File(lfn="/this/is/a/lfnB", size=1024, events=10)
        testFileA.create()
        testFileB.create()

        testJob = Job(name="TestJob", files=[testFileA, testFileB])

        assert testJob.exists() is False, \
            "ERROR: Job exists before it was created"

        testJob.create(group=testJobGroup)

        assert testJob.exists() >= 0, \
            "ERROR: Job does not exist after it was created"

        myThread = threading.currentThread()
        myThread.transaction.begin()

        testJob.delete()

        assert testJob.exists() is False, \
            "ERROR: Job exists after it was delete"

        myThread.transaction.rollback()

        assert testJob.exists() >= 0, \
            "ERROR: Job does not exist after transaction was rolled back."

        return

    def testCreateDeleteExistsNoFiles(self):
        """
        _testCreateDeleteExistsNoFiles_

        Create and then delete a job but don't add any input files to it.
        Use the job class's exists() method to determine if the job has been
        written to the database before it is created, after it has been created
        and after it has been deleted.
        """
        testWorkflow = Workflow(spec="spec.xml", owner="Simon",
                                name="wf001", task="Test")
        testWorkflow.create()

        testWMBSFileset = Fileset(name="TestFileset")
        testWMBSFileset.create()

        testSubscription = Subscription(fileset=testWMBSFileset,
                                        workflow=testWorkflow)
        testSubscription.create()

        testJobGroup = JobGroup(subscription=testSubscription)
        testJobGroup.create()

        testJob = Job(name="TestJob")

        assert testJob.exists() is False, \
            "ERROR: Job exists before it was created"

        testJob.create(group=testJobGroup)

        assert testJob.exists() >= 0, \
            "ERROR: Job does not exist after it was created"

        testJob.delete()

        assert testJob.exists() is False, \
            "ERROR: Job exists after it was delete"

        return

    def testLoad(self):
        """
        _testLoad_

        Create a job and save it to the database.  Load it back from the
        database using the name and the id and then verify that all information
        was loaded correctly.
        """
        testJobA = self.createTestJob()
        testJobB = Job(id=testJobA["id"])
        testJobC = Job(name=testJobA["name"])
        testJobB.load()
        testJobC.load()

        assert (testJobA["id"] == testJobB["id"]) and \
               (testJobA["name"] == testJobB["name"]) and \
               (testJobA["jobgroup"] == testJobB["jobgroup"]) and \
               (testJobA["couch_record"] == testJobB["couch_record"]) and \
               (testJobA["location"] == testJobB["location"]), \
            "ERROR: Load from ID didn't load everything correctly"

        assert (testJobA["id"] == testJobC["id"]) and \
               (testJobA["name"] == testJobC["name"]) and \
               (testJobA["jobgroup"] == testJobC["jobgroup"]) and \
               (testJobA["couch_record"] == testJobC["couch_record"]) and \
               (testJobA["location"] == testJobC["location"]), \
            "ERROR: Load from name didn't load everything correctly"

        self.assertEqual(testJobB['outcome'], 'failure')
        self.assertEqual(testJobC['outcome'], 'failure')
        self.assertEqual(testJobB['fwjr'], None)
        self.assertEqual(testJobC['fwjr'], None)

        return

    def testLoadData(self):
        """
        _testLoadData_

        Create a job and save it to the database.  Load it back from the
        database using the name and the id.  Verify that all job information
        is correct including input files and the job mask.
        """
        testJobA = self.createTestJob()

        testJobA["mask"]["FirstEvent"] = 1
        testJobA["mask"]["LastEvent"] = 2
        testJobA["mask"]["FirstLumi"] = 3
        testJobA["mask"]["LastLumi"] = 4
        testJobA["mask"]["FirstRun"] = 5
        testJobA["mask"]["LastRun"] = 6

        testJobA.save()

        testJobB = Job(id=testJobA["id"])
        testJobC = Job(name=testJobA["name"])
        testJobB.loadData()
        testJobC.loadData()

        assert (testJobA["id"] == testJobB["id"]) and \
               (testJobA["name"] == testJobB["name"]) and \
               (testJobA["jobgroup"] == testJobB["jobgroup"]) and \
               (testJobA["couch_record"] == testJobB["couch_record"]) and \
               (testJobA["location"] == testJobB["location"]), \
            "ERROR: Load from ID didn't load everything correctly"

        assert (testJobA["id"] == testJobC["id"]) and \
               (testJobA["name"] == testJobC["name"]) and \
               (testJobA["jobgroup"] == testJobC["jobgroup"]) and \
               (testJobA["couch_record"] == testJobC["couch_record"]) and \
               (testJobA["location"] == testJobC["location"]), \
            "ERROR: Load from name didn't load everything correctly"

        assert testJobA["mask"] == testJobB["mask"], \
            "ERROR: Job mask did not load properly"

        assert testJobA["mask"] == testJobC["mask"], \
            "ERROR: Job mask did not load properly"

        goldenFiles = testJobA.getFiles()
        for testFile in testJobB.getFiles():
            assert testFile in goldenFiles, \
                "ERROR: Job loaded an unknown file"
            goldenFiles.remove(testFile)

        assert not goldenFiles, "ERROR: Job didn't load all files"

        goldenFiles = testJobA.getFiles()
        for testFile in testJobC.getFiles():
            assert testFile in goldenFiles, \
                "ERROR: Job loaded an unknown file"
            goldenFiles.remove(testFile)

        assert not goldenFiles, "ERROR: Job didn't load all files"

        return

    def testGetFiles(self):
        """
        _testGetFiles_

        Test the Job's getFiles() method.  This should load the files from
        the database if they haven't been loaded already.
        """
        testJobA = self.createTestJob()

        testJobB = Job(id=testJobA["id"])
        testJobB.loadData()

        goldenFiles = testJobA.getFiles()
        for testFile in testJobB.getFiles():
            assert testFile in goldenFiles, \
                "ERROR: Job loaded an unknown file: %s" % testFile
            goldenFiles.remove(testFile)

        assert not goldenFiles, "ERROR: Job didn't load all files"

        return

    def testSaveTransaction(self):
        """
        _testSaveTransaction_

        Create a job and a job mask and save them both to the database.  Load
        the job from the database and verify that everything was written
        correctly.  Begin a new transaction and update the job mask again.
        Load the mask and verify that it's correct.  Finally, rollback the
        transaction and reload the mask to verify that it is in the correct
        state.
        """
        testJobA = self.createTestJob()

        testJobA["mask"]["FirstEvent"] = 1
        testJobA["mask"]["LastEvent"] = 2
        testJobA["mask"]["FirstLumi"] = 3
        testJobA["mask"]["LastLumi"] = 4
        testJobA["mask"]["FirstRun"] = 5
        testJobA["mask"]["LastRun"] = 6

        testJobA.save()

        testJobB = Job(id=testJobA["id"])
        testJobB.loadData()

        assert testJobA["mask"] == testJobB["mask"], \
            "ERROR: Job mask did not load properly"

        myThread = threading.currentThread()
        myThread.transaction.begin()

        testJobA["mask"]["FirstEvent"] = 7
        testJobA["mask"]["LastEvent"] = 8
        testJobA["mask"]["FirstLumi"] = 9
        testJobA["mask"]["LastLumi"] = 10
        testJobA["mask"]["FirstRun"] = 11
        testJobA["mask"]["LastRun"] = 12
        testJobA["name"] = "stevesJob"
        testJobA["couch_record"] = "someCouchRecord"
        testJobA["location"] = "test2.site.ch"

        testJobA.save()
        testJobC = Job(id=testJobA["id"])
        testJobC.loadData()

        assert testJobA["mask"] == testJobC["mask"], \
            "ERROR: Job mask did not load properly"

        assert testJobC["name"] == "stevesJob", \
            "ERROR: Job name did not save"

        assert testJobC["couch_record"] == "someCouchRecord", \
            "ERROR: Job couch record did not save"

        assert testJobC["location"] == "test2.site.ch", \
            "ERROR: Job site did not save"

        myThread.transaction.rollback()

        testJobD = Job(id=testJobA["id"])
        testJobD.loadData()

        assert testJobB["mask"] == testJobD["mask"], \
            "ERROR: Job mask did not load properly"

        return

    def testJobState(self):
        """
        _testJobState_

        Unittest to see if we can figure out what the jobState actually is and set it
        """

        testJobA = self.createTestJob()

        value = testJobA.getState()

        self.assertEqual(value, 'new')

        return

    def testJobCacheDir(self):
        """
        _testJobCacheDir_

        Check retrieval of the jobCache directory.
        """
        testJobA = self.createTestJob()
        value = testJobA.getCache()

        self.assertEqual(value, None)

        testJobA.setCache('UnderTheDeepBlueSea')
        value = testJobA.getCache()

        self.assertEqual(value, 'UnderTheDeepBlueSea')

        return

    def testGetOutputParentLFNs(self):
        """
        _testGetOutputParentLFNs_

        Verify that the getOutputDBSParentLFNs() method returns the correct
        parent LFNs.
        """
        testWorkflow = Workflow(spec="spec.xml", owner="Simon",
                                name="wf001", task="Test")
        testWorkflow.create()

        testWMBSFileset = Fileset(name="TestFileset")
        testWMBSFileset.create()

        testSubscription = Subscription(fileset=testWMBSFileset,
                                        workflow=testWorkflow)
        testSubscription.create()

        testJobGroup = JobGroup(subscription=testSubscription)
        testJobGroup.create()

        testFileA = File(lfn="/this/is/a/lfnA", size=1024, events=10,
                         merged=True)
        testFileB = File(lfn="/this/is/a/lfnB", size=1024, events=10,
                         merged=True)
        testFileC = File(lfn="/this/is/a/lfnC", size=1024, events=10,
                         merged=False)
        testFileD = File(lfn="/this/is/a/lfnD", size=1024, events=10,
                         merged=False)
        testFileE = File(lfn="/this/is/a/lfnE", size=1024, events=10,
                         merged=True)
        testFileF = File(lfn="/this/is/a/lfnF", size=1024, events=10,
                         merged=True)
        testFileA.create()
        testFileB.create()
        testFileC.create()
        testFileD.create()
        testFileE.create()
        testFileF.create()

        testFileE.addChild(testFileC["lfn"])
        testFileF.addChild(testFileD["lfn"])

        testJobA = Job(name="TestJob", files=[testFileA, testFileB])
        testJobA["couch_record"] = "somecouchrecord"
        testJobA["location"] = "test.site.ch"
        testJobA.create(group=testJobGroup)
        testJobA.associateFiles()

        testJobB = Job(name="TestJobB", files=[testFileC, testFileD])
        testJobB["couch_record"] = "somecouchrecord"
        testJobB["location"] = "test.site.ch"
        testJobB.create(group=testJobGroup)
        testJobB.associateFiles()

        goldenLFNs = ["/this/is/a/lfnA", "/this/is/a/lfnB"]

        parentLFNs = testJobA.getOutputDBSParentLFNs()
        for parentLFN in parentLFNs:
            assert parentLFN in goldenLFNs, \
                "ERROR: Unknown lfn: %s" % parentLFN
            goldenLFNs.remove(parentLFN)

        assert not goldenLFNs, "ERROR: LFNs are missing: %s" % goldenLFNs

        goldenLFNs = ["/this/is/a/lfnE", "/this/is/a/lfnF"]

        parentLFNs = testJobB.getOutputDBSParentLFNs()
        for parentLFN in parentLFNs:
            assert parentLFN in goldenLFNs, \
                "ERROR: Unknown lfn: %s" % parentLFN
            goldenLFNs.remove(parentLFN)

        assert not goldenLFNs, "ERROR: LFNs are missing..."

        return

    def testJobFWJRPath(self):
        """
        _testJobFWJRPath_

        Verify the correct operation of the Jobs.SetFWJRPath and
        Jobs.GetFWJRByState DAOs.
        """
        testJobA = self.createTestJob()
        testJobA["state"] = "complete"
        testJobB = self.createTestJob()
        testJobB["state"] = "executing"
        testJobC = self.createTestJob()
        testJobC["state"] = "complete"

        myThread = threading.currentThread()
        setFWJRAction = self.daoFactory(classname="Jobs.SetFWJRPath")
        setFWJRAction.execute(jobID=testJobA["id"], fwjrPath="NonsenseA",
                              conn=myThread.transaction.conn,
                              transaction=True)
        setFWJRAction.execute(jobID=testJobB["id"], fwjrPath="NonsenseB",
                              conn=myThread.transaction.conn,
                              transaction=True)
        setFWJRAction.execute(jobID=testJobC["id"], fwjrPath="NonsenseC",
                              conn=myThread.transaction.conn,
                              transaction=True)

        changeStateAction = self.daoFactory(classname="Jobs.ChangeState")
        changeStateAction.execute(jobs=[testJobA, testJobB, testJobC],
                                  conn=myThread.transaction.conn,
                                  transaction=True)

        getJobsAction = self.daoFactory(classname="Jobs.GetFWJRByState")
        jobs = getJobsAction.execute(state="complete",
                                     conn=myThread.transaction.conn,
                                     transaction=True)

        goldenIDs = [testJobA["id"], testJobC["id"]]
        for job in jobs:
            assert job["id"] in goldenIDs, \
                "Error: Unknown job: %s" % job["id"]

            goldenIDs.remove(job["id"])

            if job["id"] == testJobA["id"]:
                assert job["fwjr_path"] == "NonsenseA", \
                    "Error: Wrong fwjr path: %s" % job["fwjr_path"]
            else:
                assert job["fwjr_path"] == "NonsenseC", \
                    "Error: Wrong fwjr path: %s" % job["fwjr_path"]

        assert not goldenIDs, "Error: Jobs missing: %s" % len(goldenIDs)

        return

    def testFailJobInput(self):
        """
        _testFailJobInput_

        Test the Jobs.FailInput DAO and verify that it doesn't affect other
        jobs/subscriptions that run over the same files.
        """
        testWorkflow = Workflow(spec="spec.xml", owner="Steve",
                                name="wf001", task="Test")
        bogusWorkflow = Workflow(spec="spec1.xml", owner="Steve",
                                 name="wf002", task="Test")
        testWorkflow.create()
        bogusWorkflow.create()

        testFileset = Fileset(name="TestFileset")
        bogusFileset = Fileset(name="BogusFileset")
        testFileset.create()
        bogusFileset.create()

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=testWorkflow)
        bogusSubscription = Subscription(fileset=bogusFileset,
                                         workflow=bogusWorkflow)
        testSubscription.create()
        bogusSubscription.create()

        testFileA = File(lfn=makeUUID(), locations="T2_CH_CERN")
        testFileB = File(lfn=makeUUID(), locations="T2_CH_CERN")
        testFileC = File(lfn=makeUUID(), locations="T2_CH_CERN")
        testFileA.create()
        testFileB.create()
        testFileC.create()

        testFileset.addFile([testFileA, testFileB, testFileC])
        bogusFileset.addFile([testFileA, testFileB, testFileC])
        testFileset.commit()
        bogusFileset.commit()

        testSubscription.completeFiles([testFileA, testFileB, testFileC])
        bogusSubscription.acquireFiles([testFileA, testFileB, testFileC])

        testJobGroup = JobGroup(subscription=testSubscription)
        bogusJobGroup = JobGroup(subscription=bogusSubscription)
        testJobGroup.create()
        bogusJobGroup.create()

        testJobA = Job(name="TestJobA", files=[testFileA, testFileB, testFileC])
        testJobB = Job(name="TestJobB", files=[testFileA, testFileB, testFileC])

        bogusJob = Job(name="BogusJob", files=[testFileA, testFileB, testFileC])

        testJobA.create(group=testJobGroup)
        testJobB.create(group=testJobGroup)

        bogusJob.create(group=bogusJobGroup)

        testJobA.failInputFiles()
        testJobB.failInputFiles()

        self.assertEqual(len(testSubscription.filesOfStatus("Available")), 0)
        self.assertEqual(len(testSubscription.filesOfStatus("Acquired")), 0)
        self.assertEqual(len(testSubscription.filesOfStatus("Failed")), 3)
        self.assertEqual(len(testSubscription.filesOfStatus("Completed")), 0)

        changeStateAction = self.daoFactory(classname="Jobs.ChangeState")
        testJobB["state"] = "cleanout"
        changeStateAction.execute([testJobB])

        # Try again

        testJobA.failInputFiles()

        # Should now be failed
        self.assertEqual(len(testSubscription.filesOfStatus("Available")), 0)
        self.assertEqual(len(testSubscription.filesOfStatus("Acquired")), 0)
        self.assertEqual(len(testSubscription.filesOfStatus("Failed")), 3)
        self.assertEqual(len(testSubscription.filesOfStatus("Completed")), 0)

        # bogus should be unchanged
        self.assertEqual(len(bogusSubscription.filesOfStatus("Available")), 0)
        self.assertEqual(len(bogusSubscription.filesOfStatus("Acquired")), 3)
        self.assertEqual(len(bogusSubscription.filesOfStatus("Failed")), 0)
        self.assertEqual(len(bogusSubscription.filesOfStatus("Completed")), 0)

        return

    def testCompleteJobInput(self):
        """
        _testCompleteJobInput_

        Verify the correct output of the CompleteInput DAO.  This should mark
        the input for a job as complete once all the jobs that run over a
        particular file have complete successfully.
        """
        testWorkflow = Workflow(spec="spec.xml", owner="Steve",
                                name="wf001", task="Test")
        bogusWorkflow = Workflow(spec="spec1.xml", owner="Steve",
                                 name="wf002", task="Test")
        testWorkflow.create()
        bogusWorkflow.create()

        testFileset = Fileset(name="TestFileset")
        bogusFileset = Fileset(name="BogusFileset")
        testFileset.create()
        bogusFileset.create()

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=testWorkflow)
        bogusSubscription = Subscription(fileset=bogusFileset,
                                         workflow=bogusWorkflow)
        testSubscription.create()
        bogusSubscription.create()

        testFileA = File(lfn=makeUUID(), locations="T2_CH_CERN")
        testFileB = File(lfn=makeUUID(), locations="T2_CH_CERN")
        testFileC = File(lfn=makeUUID(), locations="T2_CH_CERN")
        testFileA.create()
        testFileB.create()
        testFileC.create()

        testFileset.addFile([testFileA, testFileB, testFileC])
        bogusFileset.addFile([testFileA, testFileB, testFileC])
        testFileset.commit()
        bogusFileset.commit()

        testSubscription.acquireFiles([testFileA, testFileB, testFileC])
        bogusSubscription.acquireFiles([testFileA, testFileB, testFileC])

        testJobGroup = JobGroup(subscription=testSubscription)
        bogusJobGroup = JobGroup(subscription=bogusSubscription)
        testJobGroup.create()
        bogusJobGroup.create()

        testJobA = Job(name="TestJobA", files=[testFileA])
        testJobB = Job(name="TestJobB", files=[testFileA, testFileB])
        testJobC = Job(name="TestJobC", files=[testFileC])
        bogusJob = Job(name="BogusJob", files=[testFileA, testFileB, testFileC])
        testJobA.create(group=testJobGroup)
        testJobB.create(group=testJobGroup)
        testJobC.create(group=testJobGroup)
        bogusJob.create(group=bogusJobGroup)

        testJobA["outcome"] = "success"
        testJobB["outcome"] = "failure"
        testJobC["outcome"] = "success"
        testJobA.save()
        testJobB.save()
        testJobC.save()

        testJobA.completeInputFiles()

        compFiles = len(testSubscription.filesOfStatus("Completed"))
        assert compFiles == 0, \
            "Error: test sub has wrong number of complete files: %s" % compFiles

        testJobB["outcome"] = "success"
        testJobB.save()

        testJobB.completeInputFiles(skipFiles=[testFileB["lfn"]])

        availFiles = len(testSubscription.filesOfStatus("Available"))
        assert availFiles == 0, \
            "Error: test sub has wrong number of available files: %s" % availFiles

        acqFiles = len(testSubscription.filesOfStatus("Acquired"))
        assert acqFiles == 1, \
            "Error: test sub has wrong number of acquired files: %s" % acqFiles

        compFiles = len(testSubscription.filesOfStatus("Completed"))
        assert compFiles == 1, \
            "Error: test sub has wrong number of complete files: %s" % compFiles

        failFiles = len(testSubscription.filesOfStatus("Failed"))
        assert failFiles == 1, \
            "Error: test sub has wrong number of failed files: %s" % failFiles

        availFiles = len(bogusSubscription.filesOfStatus("Available"))
        assert availFiles == 0, \
            "Error: test sub has wrong number of available files: %s" % availFiles

        acqFiles = len(bogusSubscription.filesOfStatus("Acquired"))
        assert acqFiles == 3, \
            "Error: test sub has wrong number of acquired files: %s" % acqFiles

        compFiles = len(bogusSubscription.filesOfStatus("Completed"))
        assert compFiles == 0, \
            "Error: test sub has wrong number of complete files: %s" % compFiles

        failFiles = len(bogusSubscription.filesOfStatus("Failed"))
        assert failFiles == 0, \
            "Error: test sub has wrong number of failed files: %s" % failFiles

        return

    def testJobTypeDAO(self):
        """
        _testJobTypeDAO_

        Verify that the Jobs.GetType DAO returns the correct job type.  The
        job type is retrieved from the subscription type. When only a single
        job is passed.
        """
        testJob = self.createTestJob()

        jobTypeAction = self.daoFactory(classname="Jobs.GetType")
        jobType = jobTypeAction.execute(jobID=testJob["id"])

        assert jobType == "Merge", \
            "Error: GetJobType DAO returned the wrong job type."

        return

    def testJobTypeDAOBulk(self):
        """
        _testJobTypeDAOBulk_

        Verify that the Jobs.GetType DAO returns the correct job type.  The
        job type is retrieved from the subscription type. When a list of jobs
        ids is passed.
        """
        testJobA = self.createTestJob(subscriptionType="Merge")
        testJobB = self.createTestJob(subscriptionType="Processing")
        testJobC = self.createTestJob(subscriptionType="Production")
        testJobD = self.createTestJob(subscriptionType="Merge")
        testJobE = self.createTestJob(subscriptionType="Skim")

        jobIds = []
        jobIds.append(testJobA["id"])
        jobIds.append(testJobB["id"])
        jobIds.append(testJobC["id"])
        jobIds.append(testJobD["id"])
        jobIds.append(testJobE["id"])

        jobTypeAction = self.daoFactory(classname="Jobs.GetType")
        jobTypes = jobTypeAction.execute(jobID=jobIds)

        entryMap = {}
        for entry in jobTypes:
            entryMap[entry["id"]] = entry["type"]

        assert entryMap[testJobA["id"]] == "Merge", \
            "Error: GetJobType DAO returned the wrong job type."
        assert entryMap[testJobB["id"]] == "Processing", \
            "Error: GetJobType DAO returned the wrong job type."
        assert entryMap[testJobC["id"]] == "Production", \
            "Error: GetJobType DAO returned the wrong job type."
        assert entryMap[testJobD["id"]] == "Merge", \
            "Error: GetJobType DAO returned the wrong job type."
        assert entryMap[testJobE["id"]] == "Skim", \
            "Error: GetJobType DAO returned the wrong job type."

        return

    def testGetOutputMapDAO(self):
        """
        _testGetOutputMapDAO_

        Verify the proper behavior of the GetOutputMapDAO for a variety of
        different processing chains.
        """
        recoOutputFileset = Fileset(name="RECO")
        recoOutputFileset.create()
        mergedRecoOutputFileset = Fileset(name="MergedRECO")
        mergedRecoOutputFileset.create()
        alcaOutputFileset = Fileset(name="ALCA")
        alcaOutputFileset.create()
        mergedAlcaOutputFileset = Fileset(name="MergedALCA")
        mergedAlcaOutputFileset.create()
        dqmOutputFileset = Fileset(name="DQM")
        dqmOutputFileset.create()
        mergedDqmOutputFileset = Fileset(name="MergedDQM")
        mergedDqmOutputFileset.create()
        cleanupFileset = Fileset(name="Cleanup")
        cleanupFileset.create()

        testWorkflow = Workflow(spec="wf001.xml", owner="Steve",
                                name="TestWF", task="None")
        testWorkflow.create()
        testWorkflow.addOutput("output", recoOutputFileset,
                               mergedRecoOutputFileset)
        testWorkflow.addOutput("ALCARECOStreamCombined", alcaOutputFileset,
                               mergedAlcaOutputFileset)
        testWorkflow.addOutput("DQM", dqmOutputFileset,
                               mergedDqmOutputFileset)
        testWorkflow.addOutput("output", cleanupFileset)
        testWorkflow.addOutput("ALCARECOStreamCombined", cleanupFileset)
        testWorkflow.addOutput("DQM", cleanupFileset)

        testRecoMergeWorkflow = Workflow(spec="wf002.xml", owner="Steve",
                                         name="TestRecoMergeWF", task="None")
        testRecoMergeWorkflow.create()
        testRecoMergeWorkflow.addOutput("anything", mergedRecoOutputFileset,
                                        mergedRecoOutputFileset)

        testRecoProcWorkflow = Workflow(spec="wf004.xml", owner="Steve",
                                        name="TestRecoProcWF", task="None")
        testRecoProcWorkflow.create()

        testAlcaChildWorkflow = Workflow(spec="wf003.xml", owner="Steve",
                                         name="TestAlcaChildWF", task="None")
        testAlcaChildWorkflow.create()

        inputFile = File(lfn="/path/to/some/lfn", size=600000, events=60000,
                         locations="cmssrm.fnal.gov")
        inputFile.create()

        testFileset = Fileset(name="TestFileset")
        testFileset.create()
        testFileset.addFile(inputFile)
        testFileset.commit()

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=testWorkflow,
                                        split_algo="EventBased",
                                        type="Processing")

        testMergeRecoSubscription = Subscription(fileset=recoOutputFileset,
                                                 workflow=testRecoMergeWorkflow,
                                                 split_algo="WMBSMergeBySize",
                                                 type="Merge")
        testProcRecoSubscription = Subscription(fileset=recoOutputFileset,
                                                workflow=testRecoProcWorkflow,
                                                split_algo="FileBased",
                                                type="Processing")

        testChildAlcaSubscription = Subscription(fileset=alcaOutputFileset,
                                                 workflow=testAlcaChildWorkflow,
                                                 split_algo="FileBased",
                                                 type="Processing")
        testSubscription.create()
        testMergeRecoSubscription.create()
        testProcRecoSubscription.create()
        testChildAlcaSubscription.create()
        testSubscription.acquireFiles()

        testJobGroup = JobGroup(subscription=testSubscription)
        testJobGroup.create()

        testJob = Job(name="SplitJobA", files=[inputFile])
        testJob.create(group=testJobGroup)
        testJob["state"] = "complete"
        testJob.save()

        outputMapAction = self.daoFactory(classname="Jobs.GetOutputMap")
        outputMap = outputMapAction.execute(jobID=testJob["id"])

        assert len(outputMap) == 3, \
            "Error: Wrong number of outputs for primary workflow."

        goldenMap = {"output": (recoOutputFileset.id,
                                mergedRecoOutputFileset.id),
                     "ALCARECOStreamCombined": (alcaOutputFileset.id,
                                                mergedAlcaOutputFileset.id),
                     "DQM": (dqmOutputFileset.id,
                             mergedDqmOutputFileset.id)}

        for outputID in outputMap:
            for outputFilesets in outputMap[outputID]:
                if outputFilesets["merged_output_fileset"] is None:
                    self.assertEqual(outputFilesets["output_fileset"],
                                     cleanupFileset.id,
                                     "Error: Cleanup fileset is wrong.")
                    continue

                self.assertTrue(outputID in goldenMap,
                                "Error: Output identifier is missing.")
                self.assertEqual(outputFilesets["output_fileset"],
                                 goldenMap[outputID][0],
                                 "Error: Output fileset is wrong.")
                self.assertEqual(outputFilesets["merged_output_fileset"],
                                 goldenMap[outputID][1],
                                 "Error: Merged output fileset is wrong.")
                del goldenMap[outputID]

        self.assertEqual(len(goldenMap), 0,
                         "Error: Missing output maps.")

        return

    def testLocations(self):
        """
        _testLocations_

        Test setting and getting locations using DAO objects
        """
        testJob = self.createTestJob()

        jobGetLocation = self.daoFactory(classname="Jobs.GetLocation")
        jobSetLocation = self.daoFactory(classname="Jobs.SetLocation")

        result = jobGetLocation.execute(jobid=testJob['id'])
        self.assertEqual(result, [['test.site.ch']])
        jobSetLocation.execute(jobid=testJob['id'], location="test2.site.ch")
        result = jobGetLocation.execute(jobid=testJob['id'])
        self.assertEqual(result, [['test2.site.ch']])

        testJob2 = self.createTestJob()
        testJob3 = self.createTestJob()

        binds = [{'jobid': testJob['id']}, {'jobid': testJob2['id']}, {'jobid': testJob3['id']}]
        result = jobGetLocation.execute(jobid=binds)
        self.assertEqual(result,
                         [{'site_name': 'test2.site.ch', 'id': 1},
                          {'site_name': 'test.site.ch', 'id': 2},
                          {'site_name': 'test.site.ch', 'id': 3}])

        return

    def testGetDataStructsJob(self):
        """
        _testGetDataStructsJob_

        Test the ability to 'cast' as a DataStructs job type
        """
        testJob = self.createTestJob()
        testJob['test'] = 'ThisIsATest'
        testJob.baggage.section_('TestSection')
        testJob.baggage.TestSection.test = 100
        finalJob = testJob.getDataStructsJob()

        for key in finalJob:
            if key == 'input_files':
                for inputFile in testJob['input_files']:
                    self.assertEqual(inputFile.returnDataStructsFile() in finalJob['input_files'], True)
                continue
            self.assertEqual(testJob[key], finalJob[key])

        self.assertEqual(str(finalJob.__class__), "<class 'WMCore.DataStructs.Job.Job'>")
        self.assertEqual(str(finalJob["mask"].__class__), "<class 'WMCore.DataStructs.Mask.Mask'>")

        for key in testJob["mask"]:
            self.assertEqual(testJob["mask"][key], finalJob["mask"][key],
                             "Error: The masks should be the same")

        self.assertEqual(finalJob.baggage.TestSection.test, 100)
        return

    def testLoadOutputID(self):
        """
        _testLoadOutputID_

        Test whether we can load an output ID for a job
        """

        testWorkflow = Workflow(spec="spec.xml", owner="Steve",
                                name="wf001", task="Test")

        testWorkflow.create()

        testFileset = Fileset(name="TestFileset")
        testFileset.create()

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=testWorkflow)

        testSubscription.create()

        testFileA = File(lfn=makeUUID(), locations="test.site.ch")
        testFileB = File(lfn=makeUUID(), locations="test.site.ch")
        testFileA.create()
        testFileB.create()

        testFileset.addFile([testFileA, testFileB])
        testFileset.commit()

        testSubscription.acquireFiles([testFileA, testFileB])

        testJobGroup = JobGroup(subscription=testSubscription)
        testJobGroup.create()

        testJob = Job()
        testJob.create(group=testJobGroup)

        self.assertEqual(testJob.loadOutputID(), testJobGroup.output.id)

        return

    def testLoadForTaskArchiver(self):
        """
        _testLoadForTaskArchiver_

        Tests the return of the DAO for the TaskArchiver
        """
        # Create 2 jobs
        jobA = self.createTestJob()
        jobB = self.createTestJob()

        # Put a mask in one
        mask = Mask()
        mask.addRunAndLumis(1, [45])
        mask.save(jobA['id'])

        # Execute the DAO
        taskArchiverDAO = self.daoFactory(classname="Jobs.LoadForTaskArchiver")
        jobs = taskArchiverDAO.execute([jobA['id'], jobB['id']])

        # Sort the jobs and check the results, we care about id, input files and mask
        jobs.sort(key=lambda x: x['id'])

        jobAprime = jobs[0]
        lfns = [x['lfn'] for x in jobAprime['input_files']]
        self.assertTrue('/this/is/a/lfnA' in lfns, 'Missing LFN lfnA from the input files')
        self.assertTrue('/this/is/a/lfnB' in lfns, 'Missing LFN lfnB from the input files')

        for inFile in jobAprime['input_files']:
            if inFile['lfn'] == '/this/is/a/lfnA':
                run = inFile['runs'].pop()
                self.assertEqual(run.run, 1, 'The run number is wrong')
                self.assertEqual(run.lumis, [45], 'The lumis are wrong')
            else:
                run = inFile['runs'].pop()
                self.assertEqual(run.run, 1, 'The run number is wrong')
                self.assertEqual(run.lumis, [46], 'The lumis are wrong')

        mask = jobAprime['mask']
        self.assertEqual(mask['runAndLumis'], {1: [[45, 45]]}, "Wrong run and lumis in mask")

        jobBprime = jobs[1]
        for inFile in jobBprime['input_files']:
            if inFile['lfn'] == '/this/is/a/lfnA':
                run = inFile['runs'].pop()
                self.assertEqual(run.run, 1, 'The run number is wrong')
                self.assertEqual(run.lumis, [45], 'The lumis are wrong')
            else:
                run = inFile['runs'].pop()
                self.assertEqual(run.run, 1, 'The run number is wrong')
                self.assertEqual(run.lumis, [46], 'The lumis are wrong')
        runs = []
        for inputFile in jobBprime['input_files']:
            runs.extend(inputFile.getRuns())
        self.assertEqual(jobBprime['mask'].filterRunLumisByMask(runs=runs), runs, "Wrong mask in jobB")

        return

    def testMask(self):
        """
        _testMask_

        Test the new mask setup
        """

        testWorkflow = Workflow(spec="spec.xml", owner="Steve",
                                name="wf001", task="Test")

        testWorkflow.create()

        testFileset = Fileset(name="TestFileset")
        testFileset.create()

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=testWorkflow)

        testSubscription.create()

        testFileA = File(lfn=makeUUID(), locations="test.site.ch")
        testFileB = File(lfn=makeUUID(), locations="test.site.ch")
        testFileA.create()
        testFileB.create()

        testFileset.addFile([testFileA, testFileB])
        testFileset.commit()

        testSubscription.acquireFiles([testFileA, testFileB])

        testJobGroup = JobGroup(subscription=testSubscription)
        testJobGroup.create()

        testJob = Job()
        testJob['mask'].addRunAndLumis(run=100, lumis=[101, 102])
        testJob['mask'].addRunAndLumis(run=200, lumis=[201, 202])
        testJob.create(group=testJobGroup)

        loadJob = Job(id=testJob.exists())
        loadJob.loadData()

        runs = loadJob['mask'].getRunAndLumis()
        self.assertEqual(len(runs), 2)
        self.assertEqual(runs[100], [[101, 102]])
        self.assertEqual(runs[200], [[201, 202]])

        bigRun = Run(100, *[101, 102, 103, 104])
        badRun = Run(300, *[1001, 1002])
        result = loadJob['mask'].filterRunLumisByMask([bigRun, badRun])

        self.assertEqual(len(result), 1)
        alteredRun = result.pop()
        self.assertEqual(alteredRun.run, 100)
        self.assertEqual(alteredRun.lumis, [101, 102])

        run0 = Run(300, *[1001, 1002])
        run1 = Run(300, *[1001, 1002])
        loadJob['mask'].filterRunLumisByMask([run0, run1])

        return

    def testAutoIncrementCheck(self):
        """
        _AutoIncrementCheck_

        Test and see whether we can find and set the auto_increment values
        """
        myThread = threading.currentThread()
        if not myThread.dialect.lower() == 'mysql':
            return

        testWorkflow = Workflow(spec="spec.xml", owner="Steve",
                                name="wf001", task="Test")

        testWorkflow.create()

        testFileset = Fileset(name="TestFileset")
        testFileset.create()

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=testWorkflow)

        testSubscription.create()

        testFileA = File(lfn=makeUUID(), locations="test.site.ch")
        testFileB = File(lfn=makeUUID(), locations="test.site.ch")
        testFileA.create()
        testFileB.create()

        testFileset.addFile([testFileA, testFileB])
        testFileset.commit()

        testSubscription.acquireFiles([testFileA, testFileB])

        testJobGroup = JobGroup(subscription=testSubscription)
        testJobGroup.create()

        incrementDAO = self.daoFactory(classname="Jobs.AutoIncrementCheck")
        incrementDAO.execute()

        testJob = Job()
        testJob.create(group=testJobGroup)
        self.assertEqual(testJob.exists(), 1)

        incrementDAO.execute()

        testJob = Job()
        testJob.create(group=testJobGroup)
        self.assertEqual(testJob.exists(), 2)

        incrementDAO.execute(input=10)

        testJob = Job()
        testJob.create(group=testJobGroup)
        self.assertEqual(testJob.exists(), 11)

        incrementDAO.execute(input=5)

        testJob = Job()
        testJob.create(group=testJobGroup)
        self.assertEqual(testJob.exists(), 12)

        return


if __name__ == "__main__":
    unittest.main()



from __future__ import (print_function, division)

import unittest
from mock import mock

from WMCore.Storage.Backends.FNALImpl import FNALImpl


class FNALImplTest(unittest.TestCase):
    def setUp(self):
        self.FNALImpl = FNALImpl()

    def testStorageMethod_local(self):
        self.assertEqual("local", self.FNALImpl.storageMethod("test"))

    def testStorageMethod_xrdcp(self):
        self.assertEqual("xrdcp", self.FNALImpl.storageMethod("root://test"))

    def testStorageMethod_srm(self):
        self.assertEqual("srm", self.FNALImpl.storageMethod("srm://test"))

    @mock.patch('WMCore.Storage.StageOutImpl.StageOutImpl.executeCommand')
    def testCreateOutputDirectory_local(self, mock_executeCommand):
        targetdir = "test"
        command = "#!/bin/sh\n"
        command += "if [ ! -e \"%s\" ]; then\n" % targetdir
        command += " mkdir -p %s\n" % targetdir
        command += "fi\n"
        self.FNALImpl.createOutputDirectory("test/file")
        mock_executeCommand.assert_called_with(command)

    @mock.patch('WMCore.Storage.StageOutImpl.StageOutImpl.executeCommand')
    @mock.patch('WMCore.Storage.Backends.FNALImpl.LCGImpl.createOutputDirectory')
    def testCreateOutputDirectory_xrdcp(self, mock_executeCommand, mock_createOutputDirectory):
        self.FNALImpl.createOutputDirectory("root://test1/test2")
        mock_executeCommand.assert_not_called()
        mock_createOutputDirectory.assert_not_called()

    @mock.patch('WMCore.Storage.Backends.FNALImpl.LCGImpl.createOutputDirectory')
    def testCreateOutputDirectory_srm(self, mock_createOutputDirectory):
        self.FNALImpl.createOutputDirectory("srm://test1/test2")
        mock_createOutputDirectory.assert_called_with("srm://test1/test2")

    @mock.patch('WMCore.Storage.Backends.FNALImpl.LCGImpl.createSourceName')
    def testCreateSourceName_srm(self, mock_createSourceName):
        self.FNALImpl.createSourceName("proto", "srm://test1/test2")
        mock_createSourceName.assert_called_with("proto", "srm://test1/test2")

    @mock.patch('WMCore.Storage.Backends.FNALImpl.LCGImpl.createSourceName')
    def testCreateSourceName_local(self, mock_createSourceName):
        self.assertEqual("local", self.FNALImpl.createSourceName("proto", "local"))
        mock_createSourceName.assert_not_called()

    def testCreateStageOutCommand_stageInTrue(self):
        self.FNALImpl.stageIn = True
        result = self.FNALImpl.createStageOutCommand("sourcePFN", "targetPFN")
        expedctedResult = "/usr/bin/xrdcp -d 0  sourcePFN  targetPFN "
        expedctedResult += """
        EXIT_STATUS=$?
        if [[ $EXIT_STATUS != 0 ]]; then
            echo "ERROR: xrdcp exited with $EXIT_STATUS"
        fi
        exit $EXIT_STATUS
        """
        self.assertEqual(expedctedResult, result)

    def testCreateStageOutCommand_stageInTrueOptions(self):
        self.FNALImpl.stageIn = True
        result = self.FNALImpl.createStageOutCommand("sourcePFN", "targetPFN", options="--test")
        expedctedResult = "/usr/bin/xrdcp -d 0  --test  sourcePFN  targetPFN "
        expedctedResult += """
        EXIT_STATUS=$?
        if [[ $EXIT_STATUS != 0 ]]; then
            echo "ERROR: xrdcp exited with $EXIT_STATUS"
        fi
        exit $EXIT_STATUS
        """
        self.assertEqual(expedctedResult, result)

    @mock.patch('WMCore.Storage.Backends.LCGImpl.LCGImpl.createStageOutCommand')
    def testCreateStageOutCommand_srm(self, mock_createStageOutCommand):
        mock_createStageOutCommand.return_value = "test"
        result = self.FNALImpl.createStageOutCommand("srm://test", "targetPFN")
        mock_createStageOutCommand.assert_called_with("srm://test", "targetPFN", None)
        self.assertEqual("test", result)
        result = self.FNALImpl.createStageOutCommand("sourcePFN", "srm://test")
        mock_createStageOutCommand.assert_called_with("sourcePFN", "srm://test", None)
        self.assertEqual("test", result)

    @mock.patch('WMCore.Storage.Backends.LCGImpl.os.stat')
    def testCreateStageOutCommand_xrdcpOptions(self, mock_stat):
        mock_stat.return_value = [0, 1, 2, 3, 4, 5, 6]
        result = self.FNALImpl.createStageOutCommand("root://test", "targetPFN", options="test")
        expectedResult = "xrdcp-old -d 0 -f "
        expectedResult += " test "
        expectedResult += " root://test "
        expectedResult += " targetPFN "
        expectedResult += """
            EXIT_STATUS=$?
            if [[ $EXIT_STATUS != 0 ]]; then
                echo "ERROR: xrdcp exited with $EXIT_STATUS"
            fi
            exit $EXIT_STATUS
            """
        self.assertEqual(expectedResult, result)

    @mock.patch('WMCore.Storage.Backends.LCGImpl.os.stat')
    def testCreateStageOutCommand_xrdcpChecksum(self, mock_stat):
        mock_stat.return_value = [0, 1, 2, 3, 4, 5, 6]
        result = self.FNALImpl.createStageOutCommand("root://test", "targetPFN", checksums={"adler32": "32"})
        expectedResult = "xrdcp-old -d 0 -f "
        expectedResult += " root://test "
        expectedResult += " targetPFN\?eos.targetsize=6\&eos.checksum=00000032 "
        expectedResult += """
            EXIT_STATUS=$?
            if [[ $EXIT_STATUS != 0 ]]; then
                echo "ERROR: xrdcp exited with $EXIT_STATUS"
            fi
            exit $EXIT_STATUS
            """
        print(expectedResult)
        print(result)
        self.assertEqual(expectedResult, result)

    def testCreateStageOutCommand_srmAndXrdcp(self):
        result = self.FNALImpl.createStageOutCommand("srm://test", "root://test")
        self.assertEqual(1, result)
        result = self.FNALImpl.createStageOutCommand("root://test", "srm://test")
        self.assertEqual(1, result)

    @mock.patch('WMCore.Storage.StageOutImpl.StageOutImpl.executeCommand')
    def testRemoveFile_xrdcp(self, mock_executeCommand):
        self.FNALImpl.removeFile("root://test")
        mock_executeCommand.assert_called_with("xrd test rm root://test")

    @mock.patch('WMCore.Storage.Backends.FNALImpl.LCGImpl.removeFile')
    def testRemoveFile_srm(self, mock_removeFile):
        self.FNALImpl.removeFile("srm://test")
        mock_removeFile.assert_called_with("srm://test")

    @mock.patch('WMCore.Storage.StageOutImpl.StageOutImpl.executeCommand')
    def testRemoveFile_local(self, mock_executeCommand):
        self.FNALImpl.removeFile("file")
        mock_executeCommand.assert_called_with("/bin/rm file")

    @mock.patch('WMCore.Storage.StageOutImpl.StageOutImpl.executeCommand')
    def testRemoveFile_localStriped(self, mock_executeCommand):
        self.FNALImpl.removeFile("cookies.fnal.gov/test")
        mock_executeCommand.assert_called_with("/bin/rm test")



#!/usr/bin/env python
"""
_GetChildren_

Oracle implementation of DBSBufferFile.GetChildren
"""




from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.GetChildren import GetChildren as MySQLGetChildren

class GetChildren(MySQLGetChildren):
    pass



"""
Main ReqMgr web page handler.

"""

from WMCore.REST.Server import RESTFrontPage

import WMCore.ReqMgr.Service.RegExp as rx


class FrontPage(RESTFrontPage):

    def __init__(self, app, config, mount):
        """
        :arg app: reference to the application object.
        :arg config: reference to the configuration.
        :arg str mount: URL mount point.

        """

        # must be in a static content directory
        frontpage = "html/ReqMgr/index.html"
        roots = \
        {
            "html":
            {
                # without repeating the 'html' here, it doesn't work
                # due to path gymnastics in WMCore.REST.Server.py
                # rather messy figuring out static content dir by
                # counting file separators and making it compatible
                # between localhost and VM running, hence the config
                # value here
                "root": "%s/html/" % config.static_content_dir,

                "rx": rx.RX_STATIC_DIR_PATH
            },
        }
        RESTFrontPage.__init__(self, app, config, mount, frontpage, roots)



"""
__ListForSubmitter__

MySQL implementation of ListForSubmitter

Created on Apr 17, 2013

@author: dballest
"""

from WMCore.Database.DBFormatter import DBFormatter

class ListForSubmitter(DBFormatter):
    """
    _ListForSubmitter_

    List the available workflows in WMBS with the workflow
    priority
    """
    sql = """SELECT wmbs_workflow.name AS name,
                    MIN(wmbs_workflow.priority) AS priority
             FROM wmbs_workflow
             GROUP BY wmbs_workflow.name"""

    def execute(self, conn = None, transaction = False):
        result = self.dbi.processData(self.sql,
                                      conn = conn, transaction = transaction)
        return self.formatDict(result)



#!/usr/bin/env python

from __future__ import (division, print_function)

endpoint = 'https://cmsweb-prod.cern.ch/dbs/prod/global/DBSReader'

# Datasets to get all blocks, array, and lumis from
datasets = ['/HighPileUp/Run2011A-v1/RAW', '/MinimumBias/ComissioningHI-v1/RAW', '/Cosmics/ComissioningHI-v1/RAW',
            '/Cosmics/ComissioningHI-PromptReco-v1/RECO',
            '/SingleElectron/StoreResults-Run2011A-WElectron-PromptSkim-v4-ALCARECO-NOLC-36cfce5a1d3f3ab4df5bd2aa0a4fa380/USER',
            '/GammaGammaToEE_Elastic_Pt15_8TeV-lpair/Summer12-START53_V7C-v1/GEN-SIM'
           ]

calls = [['listDataTiers'],
         ['listBlocks', {'block_name': '/HighPileUp/Run2011A-v1/RAW#fabf118a-cbbf-11e0-80a9-003048caaace'}],
         ['listBlocks',
          {'block_name': '/HighPileUp/Run2011A-v1/RAW#fabf118a-cbbf-11e0-80a9-003048caaace', 'detail': True}],
         ['listBlocks', {'detail': True, 'dataset': '/HighPileUp/Run2011A-v1/RAW'}],
         ['listBlocks', {'detail': False, 'dataset': '/HighPileUp/Run2011A-v1/RAW'}],
         ['listBlocks',
          {'block_name': '/Cosmics/Commissioning2015-PromptReco-v1/RECO#004ac3ba-d09e-11e4-afad-001e67ac06a0',
           'detail': True}],
         ['listBlocks',
          {'block_name': '/Cosmics/Commissioning2015-PromptReco-v1/RECO#004ac3ba-d09e-11e4-afad-001e67ac06a0'}],
         ['listBlocks',
          {'block_name': '/HighPileUp/Run2011A-v1/RAW#fabf118a-cbbf-11e0-80a9-003048caaace', 'detail': True,
           'dataset': '/HighPileUp/Run2011A-v1/RAW'}],
         ['listRuns', {'block_name': '/HighPileUp/Run2011A-v1/RAW#fabf118a-cbbf-11e0-80a9-003048caaace'}],
         ['listRuns', {'dataset': '/HighPileUp/Run2011A-v1/RAW'}],
         ['listFileArray', {'validFileOnly': 1, 'detail': True, 'dataset': '/HighPileUp/Run2011A-v1/RAW'}],
         ['listFileArray', {'dataset': '/HighPileUp/Run2011A-v1/RAW'}],
         ['listFileArray',
          {'block_name': '/Cosmics/Commissioning2015-PromptReco-v1/RECO#004ac3ba-d09e-11e4-afad-001e67ac06a0',
           'detail': True, 'validFileOnly': 1}],
         ['listFileArray',
          {'block_name': '/HighPileUp/Run2011A-v1/RAW#fabf118a-cbbf-11e0-80a9-003048caaace', 'detail': False,
           'validFileOnly': 1}],
         ['listFileArray',
          {'block_name': '/HighPileUp/Run2011A-v1/RAW#fabf118a-cbbf-11e0-80a9-003048caaace', 'detail': True,
           'validFileOnly': 1}],
         ['listFileArray', {'logical_file_name': [
             '/store/data/Commissioning2015/Cosmics/RAW/v1/000/238/545/00000/C47FDF25-2ECF-E411-A8E2-02163E011839.root'],
             'detail': True}],
         ['listFileArray', {'logical_file_name': [
             '/store/data/Commissioning2015/Cosmics/RAW/v1/000/238/545/00000/04FBE4D8-2DCF-E411-B827-02163E0124D5.root'],
             'detail': True}],
         ['listFileArray', {'logical_file_name': [
             '/store/data/Commissioning2015/Cosmics/RAW/v1/000/238/545/00000/1043E89F-2DCF-E411-9CAE-02163E013751.root'],
             'detail': True}],
         ['listFileArray', {'logical_file_name': [
             '/store/data/Commissioning2015/Cosmics/RAW/v1/000/238/545/00000/FA4E40C0-2DCF-E411-AD1A-02163E012186.root'],
             'detail': True}],
         ['listDatasets', {'primary_ds_name': 'Jet', 'data_tier_name': 'RAW'}],
         ['listDatasets', {'primary_ds_name': 'Jet', 'data_tier_name': 'RAW', 'detail': True}],
         ['listDatasets', {'primary_ds_name': 'Jet', 'data_tier_name': 'blah'}],
         ['listDatasets', {'primary_ds_name': 'blah', 'data_tier_name': 'RAW'}],
         ['listDatasets', {'dataset_access_type': '*', 'dataset': '/MET/Run2015B-05Aug2015-v1/DQMIO'}],
         ['listDatasets', {'dataset': '/NoBPTX/Run2016F-23Sep2016-v1/DQMIO', 'dataset_access_type': '*'}],
         ['listDatasets', {'dataset': '/BprimeJetToBZ_M800GeV_Tune4C_13TeV-madgraph-tauola/Fall13-POSTLS162_V1-v1/GEN-SIM', 'dataset_access_type': '*'}],
         ['listDatasets', {'dataset': '/MinBias_TuneCP5_13TeV-pythia8_pilot/RunIIFall18MiniAOD-pilot_102X_upgrade2018_realistic_v11-v1/MINIAODSIM', 'dataset_access_type': '*'}],
         ['listDatasets', {'dataset': '/JetHT/Run2022B-PromptReco-v1/MINIAOD', 'dataset_access_type': '*'}],
         ['listFileLumiArray', {'logical_file_name': [
             '/store/data/Commissioning2015/Cosmics/RAW/v1/000/238/545/00000/C47FDF25-2ECF-E411-A8E2-02163E011839.root']}],
         ['listFileLumiArray', {'logical_file_name': [
             '/store/data/Commissioning2015/Cosmics/RAW/v1/000/238/545/00000/04FBE4D8-2DCF-E411-B827-02163E0124D5.root']}],
         ['listFileLumiArray', {'logical_file_name': [
             '/store/data/Commissioning2015/Cosmics/RAW/v1/000/238/545/00000/1043E89F-2DCF-E411-9CAE-02163E013751.root']}],
         ['listFileLumiArray', {'logical_file_name': [
             '/store/data/Commissioning2015/Cosmics/RAW/v1/000/238/545/00000/FA4E40C0-2DCF-E411-AD1A-02163E012186.root']}],
         ['listFileLumis',
          {'block_name': '/HighPileUp/Run2011A-v1/RAW#fabf118a-cbbf-11e0-80a9-003048caaace', 'validFileOnly': 1}],
         ['listFileLumis',
          {'block_name': '/Cosmics/Commissioning2015-PromptReco-v1/RECO#004ac3ba-d09e-11e4-afad-001e67ac06a0',
           'validFileOnly': 1}],
         ['listFileParents',
          {'block_name': '/Cosmics/Commissioning2015-PromptReco-v1/RECO#004ac3ba-d09e-11e4-afad-001e67ac06a0'}],
         ['listFileSummaries', {'validFileOnly': 1, 'dataset': '/HighPileUp/Run2011A-v1/RAW'}],
         ['listFileSummaries',
          {'block_name': '/HighPileUp/Run2011A-v1/RAW#fabf118a-cbbf-11e0-80a9-003048caaaceasas', 'validFileOnly': 1}],
         ['listFileSummaries', {'validFileOnly': 1, 'dataset': '/HighPileUp/Run2011A-v1/RAWblah'}],
         ['listFileSummaries',
          {'block_name': '/HighPileUp/Run2011A-v1/RAW#fabf118a-cbbf-11e0-80a9-003048caaace', 'validFileOnly': 1}],
         ['listFiles', {'dataset': '/Cosmics/ComissioningHI-v1/RAW', 'run_num': 180841,
                        'lumi_list': [8, 14, 5, 22, 10, 7, 13, 18, 23, 6, 16, 2, 4, 20, 3, 15, 9, 19, 17, 11, 12, 21, 1]}],
         ['listFiles', {'block_name': '/HighPileUp/Run2011A-v1/RAW#fabf118a-cbbf-11e0-80a9-003048caaace'}],
         ['listFiles', {'block_name': '/Cosmics/ComissioningHI-PromptReco-v1/RECO#7020873e-0dcd-11e1-9b6c-003048caaace'}],
         ['listPrimaryDatasets', {'primary_ds_name': 'Jet*'}],
         ['listPrimaryDatasets', {'primary_ds_name': 'DoesntExist'}],
         ['listBlockParents',
          {'block_name': '/Cosmics/Commissioning2015-PromptReco-v1/RECO#004ac3ba-d09e-11e4-afad-001e67ac06a0'}],
         ['listBlockParents',
          {'block_name': '/Cosmics/Commissioning2015-v1/RAW#942d76fe-cf0e-11e4-afad-001e67ac06a0'}],
         ['listParentDSTrio', {'dataset': '/HToAATo4L_H5000A10_TuneCP5_13TeV-pythia8/RunIISummer20UL16MiniAODAPVv2-106X_mcRun2_asymptotic_preVFP_v11-v2/MINIAODSIM'}],
         ['listBlockTrio',
          {'block_name': '/HToAATo4L_H5000A10_TuneCP5_13TeV-pythia8/RunIISummer20UL16MiniAODAPVv2-106X_mcRun2_asymptotic_preVFP_v11-v2/MINIAODSIM#bf137a20-65c5-4927-8f62-7733c3d9e017'}],

         # Exception throwing calls

         ['listBlocks',
          {'block_name': '/HighPileUp/Run2011A-v1/RAW#fabf118a-cbbf-11e0-80a9-003048caaaceasas', 'detail': True}],
         ['listBlocks', {'block_name': '/HighPileUp/Run2011A-v1/RAW#somethingelse'}],
         ['listBlocks', {'block_name': '/HighPileUp/Run2011A-v1/RAW#fabf118a-cbbf-11e0-80a9-003048caaaceasas'}],
         ['listBlocks', {'block_name': '/HighPileUp/Run2011A-v1/RAW#fabf118a-cbbf-11e0-80a9-003048caaaceasas'}],
         ['listBlocks', {'detail': True, 'dataset': '/HighPileUp/Run2011A-v1/RAWblah'}],
         ['listBlocks', {'detail': False, 'dataset': '/thisdoesntexist/ComissioningHI-v1/RAW'}],
         ['listBlocks',
          {'block_name': '/HighPileUp/Run2011A-v1/RAW#fabf118a-cbbf-11e0-80a9-003048caaaceasas', 'detail': True,
           'dataset': '/HighPileUp/Run2011A-v1/RAW'}],
         ['listBlocks', {'block_name': '/HighPileUp/Run2011A-v1/RAW#blah'}],
         ['listBlocks', {'detail': True, 'dataset': '/thisdoesntexist/ComissioningHI-v1/RAW'}],
         ['listDatasets', {'dataset_access_type': '*', 'dataset': '/HighPileUp/Run2011A-v1/RAWblah'}],
         ['listDatasets', {'dataset_access_type': '*', 'dataset': '/thisdoesntexist/ComissioningHI-v1/RAW'}],
         ['listDatasets', {'dataset_access_type': '*', 'dataset': '/MinimumBias/FAKE-Filter-v1/RECO'}],
         ['listDatasets', {'logical_file_name': '/store/data/ComissioningHI/Cosmics/RECO/PromptReco-v1/000/180/841/368B76AA-4F09-E111-82CB-BCAEC5329721.root'}],
         ['listParentDSTrio', {'dataset': '/Cosmics/ComissioningHI-PromptReco-v1/RECO'}],
         ]



#!/usr/bin/env python
"""
_UpdateThresholdsInBulk_

This module updates thresholds for the given sites for Oracle
"""


from WMCore.ResourceControl.MySQL.UpdateThresholdsInBulk \
  import UpdateThresholdsInBulk as UpdateThresholdsInBulkMySQL
class UpdateThresholdsInBulk(UpdateThresholdsInBulkMySQL):
    pass



#!/usr/bin/env python
"""
_Delete_

Oracle implementation of Locations.Delete
"""

__all__ = []



from WMCore.WMBS.MySQL.Locations.Delete import Delete as DeleteLocationsMySQL

class Delete(DeleteLocationsMySQL):
    pass


#!/usr/bin/env python
"""
_FileTools_t_

Test class for file handling tools
"""

import os
import os.path
import unittest

from Utils import FileTools
from WMQuality.TestInitCouchApp import TestInitCouchApp


class testFileTools(unittest.TestCase):
    """
    Test to see whether we can do Linux
    """

    def setUp(self):
        """
        Do nothing

        """
        self.testInit = TestInitCouchApp(__file__)
        self.testDir = self.testInit.generateWorkDir()

        return

    def tearDown(self):
        """
        Do nothing

        """
        self.testInit.delWorkDir()

        return

    def test_tail(self):
        """
        _tail_

        Can we tail a file?
        """

        a = "a\nb\nc\nd\ne\nf\ng\nh\ni\nj\nk\nl\nm\nn\no\np\n"

        f = open('tmpfile.tmp', 'w')
        f.write(a)
        f.close()

        self.assertEqual(FileTools.tail('tmpfile.tmp', 10), "g\nh\ni\nj\nk\nl\nm\nn\no\np\n")

        self.assertEqual(FileTools.tail('tmpfile.tmp', 2), "o\np\n")

        os.remove('tmpfile.tmp')

        return

    def test_fileInfo(self):
        """
        _fileInfo_

        Test for basic file info
        """
        silly = "This is a rather ridiculous string"
        filename = os.path.join(self.testDir, 'fileInfo.test')

        with open(filename, 'w') as fObj:
            fObj.write(silly)

        info = FileTools.getFileInfo(filename=filename)
        self.assertEqual(info['Name'], filename)
        self.assertEqual(info['Size'], 34)
        return

    def test_getFullPath(self):
        fullPath = FileTools.getFullPath("cd")
        # assuming there is path for cd
        self.assertIsNotNone(fullPath)
        fullPath = FileTools.getFullPath("this_shouldnt_be")
        self.assertEqual(fullPath, None)

    def testChecksum(self):
        """
        Test file checksums calculation
        """
        filename = os.path.join(self.testDir, 'fileInfo.test')
        with open(filename, 'w') as fObj:
            fObj.write("")

        (adler32, cksum) = FileTools.calculateChecksums(filename=filename)
        self.assertEqual(adler32, "00000001")
        self.assertEqual(cksum, "4294967295")

        silly = "This is a rather ridiculous string"
        with open(filename, 'w') as fObj:
            fObj.write(silly * 100001)
        (adler32, cksum) = FileTools.calculateChecksums(filename=filename)
        self.assertEqual(adler32, "827db5b1")
        self.assertEqual(cksum, "3774692924")
        return

    def testTarMode(self):
        """
        Test tarMode function
        """
        mode = FileTools.tarMode('file.bz2', 'w')
        self.assertEqual(mode, 'w:bz2')
        mode = FileTools.tarMode('file.tar', 'r')
        self.assertEqual(mode, 'r')


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
_MathAlgos_t_

Test class for basic Math Algorithms
"""


import unittest

from WMCore.Algorithms import MathAlgos

class MathAlgoTest(unittest.TestCase):
    """
    Tests for basic Math Algorithms

    """


    def setUp(self):
        """
        Do nothing

        """
        return

    def tearDown(self):
        """
        Do nothing

        """
        return

    def testAverageStdDev(self):
        """
        _testAverageStdDev_

        Test average, standard deviation function
        """

        numList = ['a', 'b', 'c']
        self.assertRaises(MathAlgos.MathAlgoException,
                          MathAlgos.getAverageStdDev, numList)

        numList = [1, 1, 1, 1, 1, 1, 1, 1]
        result = MathAlgos.getAverageStdDev(numList = numList)
        self.assertEqual(result[0], 1.0)  # Average should be zero
        self.assertEqual(result[1], 0.0)  # stdDev should be zero

        numList = [1, 2, 3, 4, 5, 6, 7, 8]
        result = MathAlgos.getAverageStdDev(numList = numList)
        self.assertEqual(result[0], 4.5)
        self.assertEqual(result[1], 2.2912878474779199) # I think this is right

        return

    def testTruncate(self):
        """
        _testTruncate_

        Test the floorTruncate function
        """

        self.assertEqual(MathAlgos.floorTruncate(1.23456), 1.234)
        return


    def testHistogram(self):
        """
        _testHistogram_

        Test our ability to build histogram objects out of lists
        """

        # Check that we override correctly when we give a uniform list
        numList = [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
        result = MathAlgos.createHistogram(numList = numList, nBins = 10, limit = 10)
        self.assertEqual(result[0]['lowerEdge'], 0.0)
        self.assertEqual(result[0]['upperEdge'], 2.0)
        self.assertEqual(result[0]['nEvents'],   10)

        numList = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        result = MathAlgos.createHistogram(numList = numList, nBins = 2, limit = 10)
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0]['nEvents'], 5)
        self.assertEqual(result[1]['nEvents'], 5)
        self.assertEqual(result[0]['average'], 3.0)
        self.assertEqual(result[1]['average'], 8.0)
        self.assertEqual(result[0]['stdDev'],  1.4142135623730951)
        self.assertEqual(result[1]['stdDev'],  1.4142135623730951)

        # Check that we generate overflow and underflow bins correctly
        result = MathAlgos.createHistogram(numList = numList, nBins = 2, limit = 1)
        self.assertEqual(len(result), 4)
        self.assertEqual(result[0]['type'], 'underflow')
        self.assertEqual(result[1]['type'], 'overflow')
        self.assertEqual(result[0]['nEvents'], 2)
        self.assertEqual(result[1]['nEvents'], 2)
        self.assertEqual(result[0]['stdDev'], 0.5)
        self.assertEqual(result[1]['stdDev'], 0.5)


    def testSortListByKeys(self):
        """
        _testSortListByKeys_

        Test our ability to sort a list of dictionaries in the order of a single
        numerical key
        """

        l = [{'a': 102, 'b': 200, 'name': 'One'},
             {'a': 101, 'b': 199, 'name': 'Two'},
             {'a': 100, 'b': 198, 'name': 'Three'},
             {'a': 103, 'b': 197, 'name': 'Four'}]

        result = MathAlgos.sortDictionaryListByKey(dictList = l, key = 'a')
        self.assertEqual(result[0]['name'], 'Three')
        self.assertEqual(result[1]['name'], 'Two')
        self.assertEqual(result[2]['name'], 'One')
        self.assertEqual(result[3]['name'], 'Four')

        result = MathAlgos.sortDictionaryListByKey(dictList = l, key = 'b')
        self.assertEqual(result[0]['name'], 'Four')
        self.assertEqual(result[1]['name'], 'Three')
        self.assertEqual(result[2]['name'], 'Two')
        self.assertEqual(result[3]['name'], 'One')

        result = MathAlgos.sortDictionaryListByKey(dictList = l, key = 'b',
                                                   reverse = True)
        self.assertEqual(result[3]['name'], 'Four')
        self.assertEqual(result[2]['name'], 'Three')
        self.assertEqual(result[1]['name'], 'Two')
        self.assertEqual(result[0]['name'], 'One')


        # This shouldn't fail, it just should return a flat list
        result = MathAlgos.sortDictionaryListByKey(dictList = l, key = 'c',
                                                   reverse = True)


        return

    def testGetLargestValue(self):
        """
        _testGetLargestValue_

        See if we can get the largest value from a list of histograms
        for a specific key
        """

        l = [{'a': 102, 'b': 200, 'name': 'One'},
             {'a': 101, 'b': 199, 'name': 'Two'},
             {'a': 100, 'b': 198, 'name': 'Three'},
             {'a': 103, 'b': 197, 'name': 'Four'}]

        result = MathAlgos.getLargestValues(dictList = l, key = 'a', n = 2)
        self.assertEqual(result, [{'a': 103, 'b': 197, 'name': 'Four'},
                                  {'a': 102, 'b': 200, 'name': 'One'}])
        result = MathAlgos.getLargestValues(dictList = l, key = 'b', n = 3)
        self.assertEqual(result, [{'a': 102, 'b': 200, 'name': 'One'},
                                  {'a': 101, 'b': 199, 'name': 'Two'},
                                  {'a': 100, 'b': 198, 'name': 'Three'}])
        return


if __name__ == "__main__":
    unittest.main()



"""
Main ReqMgr web page handler.

"""
import re
from WMCore.REST.Server import RESTFrontPage

# path to static resources
RX_STATIC_DIR_PATH = re.compile(r"^([a-zA-Z]+/)+[-a-z0-9_]+\.(?:css|js|png|gif|html)$")

class FrontPage(RESTFrontPage):

    def __init__(self, app, config, mount):
        """
        :arg app: reference to the application object.
        :arg config: reference to the configuration.
        :arg str mount: URL mount point.

        """

        # must be in a static content directory
        frontpage = "html/WMStats/index.html"
        roots = \
        {
            "html":
            {
                # without repeating the 'html' here, it doesn't work
                # due to path gymnastics in WMCore.REST.Server.py
                # rather messy figuring out static content dir by
                # counting file separators and making it compatible
                # between localhost and VM running, hence the config
                # value here
                "root": "%s/html/" % config.static_content_dir,

                "rx": RX_STATIC_DIR_PATH
            },
            "js":
            {   "root": "%s/html/WMStats/js/" % config.static_content_dir,
                "rx": re.compile(r"^([a-zA-Z]+/)+[-a-z0-9_]+\.(?:js)$")
            },
            "css":
            {   "root": "%s/html/WMStats/css/" % config.static_content_dir,
                "rx": re.compile(r"^([a-zA-Z]+/)+[-a-z0-9_]+\.(?:css)$")
            },
            "images":
            {   "root": "%s/html/WMStats/images/" % config.static_content_dir,
                "rx": re.compile(r"^([a-zA-Z]+/)+[-a-z0-9_]+\.(?:png|gif)$")
            },
            "lib":
            {   "root": "%s/html/WMStats/lib/" % config.static_content_dir,
                "rx": RX_STATIC_DIR_PATH
            },
            "fonts":
            {   "root": "%s/html/WMStats/fonts/" % config.static_content_dir,
                "rx": RX_STATIC_DIR_PATH
            }

        }
        RESTFrontPage.__init__(self, app, config, mount, frontpage, roots)



"""
Unit tests for Service/Data.py module

Author: Valentin Kuznetsov <vkuznet [AT] gmail [DOT] com>
"""

# system modules
import unittest

# third party modules
import cherrypy

# WMCore modules
from WMCore.MicroService.Service.Data import mspileupError


class ServiceData(unittest.TestCase):
    "Unit test for Service/Data module"

    def testMSPileupError(self):
        "test mspileupError function"
        doc = {'error': 'mspileup error', 'code': 123, 'message': 'msg'}
        with self.assertRaises(cherrypy.HTTPError):
            mspileupError(doc)
        success = {}
        mspileupError(success)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_EventAwareLumiByWork_

Lumi based splitting algorithm that will chop a fileset into
a set of jobs based on lumi sections, failing jobs with too many
events in a lumi and adapting the number of lumis per job
according to the average number of events per lumi in the files.

This is adapted from EventAwareLumiBased but does its work on lumi sections directly, not files.
No correction for lumis split across files is needed, it's automatic. Makes use of the LumiList
class to simplify the code

"""

from __future__ import (division, print_function)

import operator

from future.utils import viewitems

import logging
from collections import defaultdict

from WMCore.DataStructs.LumiList import LumiList
from WMCore.DataStructs.Run import Run
from WMCore.JobSplitting.JobFactory import JobFactory
from WMCore.WMBS.File import File


class EventAwareLumiByWork(JobFactory):
    """
    Split jobs by lumis taking into account events per lumi
    """

    locations = []

    def __init__(self, package='WMCore.DataStructs', subscription=None, generators=None, limit=0):
        super(EventAwareLumiByWork, self).__init__(package, subscription, generators, limit)

        self.loadRunLumi = None  # Placeholder for DAO factory if needed
        self.collectionName = None  # Placeholder for ACDC Collection Name, if needed
        self.perfParameters = {}
        self.deterministicPU = False
        self.maxLumis = 1  # Maximum lumis seen in a job (needed for deterministic pileup only)
        self.maxEvents = 1  # Maximum events seen in a lumi (needed for deterministic pileup only)
        # TODO this might need to be configurable instead of being hardcoded
        self.defaultJobTimeLimit = 48 * 3600 # 48 hours

        # Job accumulators
        self.eventsInJob = 0
        self.jobLumis = []
        self.jobFiles = set()
        self.eventsInLumi = 0

        # Location accumulators. Using these two objects lets us automatically fix up the cases where a lumi
        # is split across two files
        self.lumisProcessed = set()  # Which run/lumi pairs already have jobs created for this location
        self.filesByLumi = {}  # Map by lumi of which files contain that lumi

    def algorithm(self, *args, **kwargs):
        """
        _algorithm_

        Split files into a number of lumis per job
        Allow a flag to determine if we split files between jobs
        """

        # TODO: This has one possible weakness. If two files (blocks) contain the same run/lumi section but are
        # TODO: split across sites two different jobs will read the same lumi

        avgEventsPerJob = int(kwargs.get('events_per_job', 5000))
        jobTimeLimit = int(kwargs.get('job_time_limit', self.defaultJobTimeLimit))
        totalEventLimit = int(kwargs.get('total_events', 0))
        splitOnFile = bool(kwargs.get('halt_job_on_file_boundaries', False))
        self.collectionName = kwargs.get('collectionName', None)
        splitOnRun = kwargs.get('splitOnRun', True)
        getParents = kwargs.get('include_parents', False)
        runWhitelist = kwargs.get('runWhitelist', [])
        runs = kwargs.get('runs', None)
        lumis = kwargs.get('lumis', None)
        self.deterministicPU = kwargs.get('deterministicPileup', False)
        self.perfParameters = kwargs.get('performance', {})

        # Calculate and add performance information
        self.timePerEvent, self.sizePerEvent, self.memoryRequirement = \
            self.getPerformanceParameters(self.perfParameters)

        if avgEventsPerJob <= 0:
            msg = "events_per_job parameter must be positive. Its value is: %d" % avgEventsPerJob
            raise RuntimeError(msg)

        # Set the lumi mask for the fileset based on ACDC or runs & lumis and/or runWhitelist
        lumiMask = LumiList()
        if self.collectionName:
            lumiMask = self.lumiListFromACDC(couchURL=kwargs.get('couchURL'), couchDB=kwargs.get('couchDB'),
                                             filesetName=kwargs.get('filesetName'), collectionName=self.collectionName)
        elif runs and lumis and runWhitelist:
            lumiMask = LumiList(wmagentFormat=(runs, lumis)) & LumiList(runs=runWhitelist)
        elif runs and lumis:
            lumiMask = LumiList(wmagentFormat=(runs, lumis))
        elif runWhitelist:
            lumiMask = LumiList(runs=runWhitelist)
        logging.debug('%s splitting with lumiMask%s%s', self.__class__.__name__, '\n' if bool(lumiMask) else ' ',
                      lumiMask)

        if self.package == 'WMCore.WMBS':
            self.loadRunLumi = self.daoFactory(classname="Files.GetBulkRunLumi")
            if self.deterministicPU:
                getJobNumber = self.daoFactory(classname="Jobs.GetNumberOfJobsPerWorkflow")
                self.nJobs = getJobNumber.execute(workflow=self.subscription.getWorkflow().id)
                logging.info('Creating jobs in DeterministicPileup mode for %s',
                             self.subscription.workflowName())

        filesByLocation = self.getFilesSortedByLocation(avgEventsPerJob)
        if not filesByLocation:
            logging.info("There are not enough events/files to be splitted. Trying again next cycle")
            return

        totalEvents = 0
        lastRun = None
        stopTask = False
        lastFile = None

        for location, filesAtLocation in viewitems(filesByLocation):
            self.newGroup()  # For each location, we need a new jobGroup
            self.eventsInJob = 0
            self.jobLumis = []
            self.jobFiles = set()
            self.lumisProcessed = set()
            if self.loadRunLumi:
                self.populateFilesFromWMBS(filesAtLocation)
            lumisByFile, eventsByLumi = self.fileLumiMaps(filesAtLocation=filesAtLocation, getParents=getParents,
                                                          lumiMask=lumiMask)
            # sort files by name, to have a more reproducible job creation
            for f in sorted(filesAtLocation, key=operator.itemgetter('lfn')):
                lfn = f['lfn']
                if lfn not in lumisByFile:
                    continue  # There are no lumis of interest in the file
                for run, lumi in lumisByFile[lfn].getLumis():
                    if (run, lumi) in self.lumisProcessed:
                        continue  # We already saw this lumi and it got included in an earlier job
                    self.eventsInLumi = eventsByLumi[run][lumi]

                    if 0 < totalEventLimit <= totalEvents:
                        stopTask = True
                        break  # Don't add this lumi to the job

                    totalEvents += self.eventsInLumi
                    self.maxEvents = max(self.maxEvents, self.eventsInLumi)
                    timePerLumi = self.eventsInLumi * self.timePerEvent
                    if timePerLumi > jobTimeLimit and len(lumisByFile[lfn].getLumis()):
                        # This lumi has too many events. Output this job and a new one with just that lumi
                        failReason = "File %s has a single lumi %s, in run %s " % (lfn, lumi, run)
                        failReason += "with too many events %d and it would take %d sec to run" \
                                      % (self.eventsInLumi, timePerLumi)
                        self.stopAndMakeJob(reason='Lumi too big', runLumi=(run, lumi),
                                            failNextJob=True, failReason=failReason)
                    elif abs(self.eventsInLumi + self.eventsInJob - avgEventsPerJob) >= abs(
                                    self.eventsInJob - avgEventsPerJob) \
                            and self.eventsInLumi > 0 and self.eventsInJob > 0:
                        # This lumi doesn't fit in this job (logic is to get as close as possible to avgEventsPerJob)
                        self.stopAndMakeJob(reason='Event limit', runLumi=(run, lumi))
                    elif splitOnRun and run != lastRun and self.eventsInJob > 0:
                        # This lumi is from a new run and we break on it
                        self.stopAndMakeJob(reason='Run change', runLumi=(run, lumi))
                    elif splitOnFile and self.filesByLumi[run][lumi][0]['lfn'] != lastFile and len(self.jobLumis):
                        # This lumi is from a new file and we break on it
                        self.stopAndMakeJob(reason='File change', runLumi=(run, lumi))
                    else:  # Keep going, add lumi to list of lumis for the job
                        for f2 in self.filesByLumi[run][lumi]:
                            self.jobFiles.add(f2)
                        self.jobLumis.append((run, lumi))
                        self.eventsInJob += self.eventsInLumi
                    lastRun = run
                    lastFile = lfn

            # Make the last job assuming its not just a failed job
            if self.eventsInJob > 0 or self.jobLumis or self.jobFiles:  # Make sure there is someting to do
                self.makeNewJobByWork(reason='End of list')
            if stopTask:
                break

        return

    def stopAndMakeJob(self, runLumi, failNextJob=False, reason=None, failReason=None):
        """
        Put out the previously accumulated info into a job and start a new job with the current lumi, events

        Args:
            runLumi: The current run and lumi
            failNextJob: The current lumi belongs in its own job too, so make two jobs
            reason: The reason for starting a new job
            failReason: The reason for making the second job

        Returns: nothing
        """

        (run, lumi) = runLumi
        # Make the job assuming its not just a failed job
        if self.eventsInJob > 0 or not failNextJob:
            self.makeNewJobByWork(reason=reason)

        # Now start on the next job
        self.eventsInJob = self.eventsInLumi  # Promote both to self
        self.jobLumis = [(run, lumi)]  # Promote both to self
        self.lumisProcessed.add((run, lumi))  # Promote to self
        self.jobFiles = set()  # Promote to self
        for f2 in self.filesByLumi[run][lumi]:  # promote filesByLumi to self
            self.jobFiles.add(f2)

        if failNextJob:  # Make a 2nd failed job
            self.makeNewJobByWork(reason=failReason, failedJob=True)
            self.eventsInJob = 0
            self.jobLumis = []
            self.jobFiles = set()

    def makeNewJobByWork(self, reason='', failedJob=False):
        """
        Make a new job given the passed in parameters.

        :param reason: Why are we making a new job (debugging only)
        :param failedJob: Make the job as already failed

        :return: nothing
        """

        events = self.eventsInJob
        lumis = self.jobLumis
        files = self.jobFiles

        self.maxLumis = max(self.maxLumis, len(lumis))

        # Transform the lumi list into something compact and usable
        lumiList = LumiList(lumis=lumis).getCompactList()
        logging.debug("Because %s new job with events: %s, lumis: %s, and files: %s",
                      reason, events, lumiList, [f['lfn'] for f in files])
        if failedJob:
            logging.debug(" This job will be made failed")
            self.newJob(failedJob=failedJob, failedReason=reason)
        else:
            self.newJob()

        self.currentJob.addResourceEstimates(jobTime=events * self.timePerEvent,
                                             disk=events * self.sizePerEvent,
                                             memory=self.memoryRequirement)
        # Add job mask information
        for run, lumiRanges in viewitems(lumiList):
            for lumiRange in lumiRanges:
                self.currentJob['mask'].addRunAndLumis(run=int(run), lumis=lumiRange)
        # Add files
        for f in files:
            self.currentJob.addFile(f)
        # Add pileup info if needed
        if self.deterministicPU:
            eventsToSkip = (self.nJobs - 1) * self.maxEvents * self.maxLumis
            logging.debug('Adding baggage to skip %s events', eventsToSkip)
            self.currentJob.addBaggageParameter("skipPileupEvents", eventsToSkip)

        return

    @staticmethod
    def countLumis(runs):
        """
        Count up the number of lumi sections in a runs object
        :param runs:
        :return count:
        """
        count = 0
        for runLumi in runs:
            count += len(runLumi.lumis)
        return count

    @staticmethod
    def lumiListFromACDC(couchURL=None, couchDB=None, filesetName=None, collectionName=None):
        """
        This is not implemented yet
        :return:
        """
        from WMCore.ACDC.DataCollectionService import DataCollectionService

        goodRunList = None
        try:
            logging.info('Creating jobs for ACDC fileset %s', filesetName)
            dcs = DataCollectionService(couchURL, couchDB)
            goodRunList = dcs.getLumilistWhitelist(collectionName, filesetName)
        except Exception as ex:
            msg = "Exception while trying to load goodRunList. "
            msg += "Refusing to create any jobs.\nDetails: %s" % ex.__str__()
            logging.exception(msg)

        return goodRunList

    def fileLumiMaps(self, filesAtLocation=None, getParents=False, lumiMask=None):
        """
        Args:
            filesAtLocation: the list of file objects at a particular location
            getParents: Attach a list of parents files to the file
            lumiMask: the lumiMask for the step. Files excluded by the lumiMask are dropped

        Returns:
            lumisByFile: LumiList for each LFN
            eventsByLumi: Estimate (for now) of how many events are in each lumi
        """
        lumisByFile = {}
        eventsByLumi = defaultdict(lambda: defaultdict(float))
        self.filesByLumi = defaultdict(lambda: defaultdict(list))

        for fileObj in filesAtLocation:
            lfn = fileObj['lfn']
            eventsInFile = fileObj['events']
            lumisInFile = self.countLumis(fileObj['runs'])
            if getParents:
                parentLFNs = self.findParent(lfn=lfn)
                for lfn in parentLFNs:
                    parent = File(lfn=lfn)
                    fileObj['parents'].add(parent)
            runsAndLumis = {str(runLumi.run): runLumi.lumis for runLumi in fileObj['runs']}
            lumiList = LumiList(runsAndLumis=runsAndLumis)
            if lumiMask:  # Apply the mask if there is one
                lumiList &= lumiMask
            if lumiList:  # Skip files with no lumis of interest
                lumisByFile[lfn] = lumiList

                for runLumi in fileObj['runs']:
                    run = runLumi.run
                    for lumi in runLumi.lumis:
                        if (runLumi.run, lumi) in lumiList:
                            self.filesByLumi[run][lumi].append(fileObj)
                            eventsByLumi[run][lumi] = int(round(eventsInFile / lumisInFile))

        return lumisByFile, eventsByLumi

    def populateFilesFromWMBS(self, filesByLocation):
        """
        Load the lumi information for files from WMBS

        Args:
            filesByLocation: the files at the location currently under consideration

        Returns: nothing
        """

        fileLumis = self.loadRunLumi.execute(files=filesByLocation)
        if not fileLumis:
            logging.warning("Empty fileLumis dict for workflow %s, subs %s.",
                            self.subscription.workflowName(), self.subscription['id'])
        for f in filesByLocation:
            lumiDict = fileLumis.get(f['id'], {})
            for run in lumiDict:
                f.addRun(run=Run(run, *lumiDict[run]))
